/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <gdk/gdkkeysyms.h>

#include <gtk/gtk.h>
#if !GTK_CHECK_VERSION(2, 4, 0)
#include "my-gtkexpander.h"
#define gtk_event_box_set_visible_window(x, y)
#define g_queue_get_length(x) (x->length)
#endif
#include <gdk-pixbuf/gdk-pixdata.h>

#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>

#include "exo/exo-ellipsized-label.h"

#include "xfmedia-internal.h"

#include "gtkxine.h"
#include "mainwin.h"
#include "playlist.h"
#include <xfmedia/xfmedia-settings.h>
#include "trayicon.h"
#include "keybindings.h"
#include "xfmedia-keybindings.h"
#include "main.h"
#include "mediamarks.h"

#define MAX_SHUFFLE_QUEUE_DEPTH 100
#define AUTOHIDE_TIMEOUT        500

gboolean
xfmedia_position_cb(gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint pos_stream, pos_time, length_time;
    gchar time_lbl[128];
    
    if(gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY
            ||gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED) == XINE_SPEED_PAUSE)
    {
        return TRUE;
    }
    
    if(!gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time))
        return TRUE;
    pos_time /= 1000;
    length_time /= 1000;
    
    if(xfmedia_settings_get_bool("/xfmedia/general/show_remaining") && length_time > 0) {
        g_snprintf(time_lbl, 128,
                "<span font_family=\"monospace\" size=\"large\" weight=\"bold\">-%02d:%02d</span>",
                (length_time-pos_time) / 60, (length_time-pos_time) % 60);
        gtk_label_set_markup(GTK_LABEL(mwin->time_label), time_lbl);
        g_snprintf(time_lbl, 128,
                "<span font_family=\"monospace\">-%02d:%02d</span>",
                (length_time-pos_time) / 60, (length_time-pos_time) % 60);
        gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label), time_lbl);
    } else {
        g_snprintf(time_lbl, 128,
                "<span font_family=\"monospace\" size=\"large\" weight=\"bold\"> %02d:%02d</span>",
                pos_time / 60, pos_time % 60);
        gtk_label_set_markup(GTK_LABEL(mwin->time_label), time_lbl);
        g_snprintf(time_lbl, 128,
                "<span font_family=\"monospace\"> %02d:%02d</span>",
                pos_time / 60, pos_time % 60);
        gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label), time_lbl);
    }
    
    /* sometimes xine fails to get the song length on the first try, so we'll
     * check it here if we have a zero length */
    if(length_time != 0 && xfmedia_settings_get_int("/xfmedia/playlist/metadata_load_impetus") <= 1) {
        GtkTreePath *path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
        if(path) {
            GtkTreeIter itr;
            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list), &itr, path)) {
                gint cur_length = 0;
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list), &itr,
                        PLAYLIST_COL_LENGTH, &cur_length, -1);
                if(cur_length != length_time || mwin->plist->cur_playing_len <= 0) {
                    gchar *len_str = g_strdup_printf("%02d:%02d", length_time / 60, length_time % 60);
                    gtk_list_store_set(mwin->plist->file_list, &itr,
                            PLAYLIST_COL_LENGTH, length_time,
                            PLAYLIST_COL_LENGTH_S, len_str, -1);
                    g_free(len_str);
                    gtk_range_set_range(GTK_RANGE(mwin->position_slider), 0, length_time);
                    gtk_range_set_range(GTK_RANGE(mwin->fs_position_slider), 0, length_time);
                    mwin->plist->cur_playing_len = length_time;
                }
            }
            gtk_tree_path_free(path);
        }
    }
    
    if(!mwin->slider_sticky && length_time > 0) {
        xfmedia_mainwin_disconnect_slider_callback(mwin);
        gtk_range_set_value(GTK_RANGE(mwin->position_slider), (gdouble)pos_time);
        gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), (gdouble)pos_time);
        xfmedia_mainwin_connect_slider_callback(mwin);
    }
    
    return TRUE;
}

void
xfmedia_position_slider_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint new_pos, pos_stream, pos_time, length_time;
    gint status, speed;
    
    status = gtk_xine_get_status(mwin->gtx);
    speed = gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED);
    
    new_pos = (gint)gtk_range_get_value(GTK_RANGE(w)) * 1000;
    DBG("seeking to %d\n", new_pos);
    if(gtk_xine_play(mwin->gtx, 0, new_pos)) {
        if(!gtk_xine_get_pos_length(mwin->gtx, &pos_stream, &pos_time, &length_time))
            return;
        DBG("  new position is %d\n", pos_time);
        xfmedia_mainwin_disconnect_slider_callback(mwin);
        pos_time /= 1000;
        DBG("setting value\n");
        gtk_range_set_value(GTK_RANGE(mwin->position_slider), pos_time);
        gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), pos_time);
        DBG("value set\n");
        /*xfmedia_mainwin_connect_slider_callback(mwin);*/
        
        if(status != XINE_STATUS_PLAY || speed == XINE_SPEED_PAUSE)
            xfmedia_mainwin_set_pause_buttons(mwin);
    }
}

gboolean
xfmedia_pos_slider_btn_press_cb(GtkWidget *w, GdkEventButton *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(evt->button == 1 || evt->button == 2)
        mwin->slider_sticky = TRUE;
    
    return FALSE;
}

gboolean
xfmedia_pos_slider_btn_rls_cb(GtkWidget *w, GdkEventButton *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(evt->button == 1 || evt->button == 2)
        mwin->slider_sticky = FALSE;
    
    return FALSE;
}

void
xfmedia_playpause_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GList *selected;
    GtkTreePath *path = NULL;
    gint speed, status;
    
    speed = gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED);
    status = gtk_xine_get_status(mwin->gtx);
    
    if(status == XINE_STATUS_PLAY && speed != XINE_SPEED_PAUSE) {
        xfmedia_mainwin_set_play_buttons(mwin);
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, XINE_SPEED_PAUSE);
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
        xfmedia_plugins_forward_signal_VOID("speed-changed");
    } else if(status == XINE_STATUS_STOP || status == XINE_STATUS_IDLE) {        
        if(mwin->plist->cur_playing) {
            path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
            if(!path) {
                gtk_tree_row_reference_free(mwin->plist->cur_playing);
                mwin->plist->cur_playing = NULL;
            }
        }
        if(!path) {
            selected = gtk_tree_selection_get_selected_rows(mwin->plist->tree_sel, NULL);
            if(!selected) {
                path = gtk_tree_path_new_first();
                gtk_tree_selection_select_path(mwin->plist->tree_sel, path);
            } else {
                path = gtk_tree_path_copy((GtkTreePath *)selected->data);
                g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
                g_list_free(selected);
            }
        }
        if(path) {
            xfmedia_playlist_play_file_at_path(mwin, path);
            gtk_tree_path_free(path);
        }
    } else if(status == XINE_STATUS_PLAY && speed == XINE_SPEED_PAUSE) {
        xfmedia_mainwin_set_pause_buttons(mwin);
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 0);
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
        xfmedia_plugins_forward_signal_VOID("speed-changed");
    }
}

void
xfmedia_stop_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint speed, status;
    
    speed = gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED);
    status = gtk_xine_get_status(mwin->gtx);
    
    if(status == XINE_STATUS_PLAY && speed != XINE_SPEED_PAUSE)
        xfmedia_mainwin_set_play_buttons(mwin);
    
    if(status == XINE_STATUS_PLAY) {
        gtk_xine_stop(mwin->gtx);
        
        xfmedia_mainwin_disconnect_slider_callback(mwin);
        gtk_range_set_value(GTK_RANGE(mwin->position_slider), 0.0);
        gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), 0.0);
        xfmedia_mainwin_connect_slider_callback(mwin);
        
        gtk_label_set_markup(GTK_LABEL(mwin->time_label),
                "<span font_family=\"monospace\" size=\"large\" weight=\"bold\"> 00:00</span>");
        gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label),
                "<span font_family=\"monospace\"> 00:00</span>");
        
        if(GTK_WIDGET_VISIBLE(mwin->video_window))
            gtk_widget_hide(mwin->video_window);
        if(gtk_xine_is_fullscreen(mwin->gtx)) {
            if(GTK_WIDGET_VISIBLE(mwin->fs_controls))
                gtk_widget_hide(mwin->fs_controls);
            gtk_xine_set_fullscreen(mwin->gtx, FALSE);
        }
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
    }
}

void
xfmedia_prev_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkTreePath *path = NULL;
    gint tot_items, *indices = NULL;
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")) {
        GtkTreeRowReference *rref = NULL;
        gboolean playing = FALSE;
        
        if(mwin->plist->cur_playing) {
            GtkTreePath *oldpath = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
            if(oldpath) {
                xfmedia_playlist_shuffle_unset(mwin->plist,
                        xfmedia_tree_path_to_index(oldpath));
                gtk_tree_path_free(oldpath);
            }
        }
        
        do {
            if(path) {
                xfmedia_playlist_shuffle_unset(mwin->plist,
                        xfmedia_tree_path_to_index(path));
                g_free(path);
                path = NULL;
            }
            rref = g_queue_pop_head(mwin->plist->shuffle_history);
            DBG("got rref from shuffle history %p", rref);
            if(rref) {
                path = gtk_tree_row_reference_get_path(rref);
                gtk_tree_row_reference_free(rref);
            }
        } while(path &&    !(playing=xfmedia_playlist_play_file_at_path(mwin, path)));
        if(path)
            gtk_tree_path_free(path);
        
        if(!playing)
            xfmedia_playlist_play_file_at_path_persistent(mwin, NULL);
    } else if(mwin->plist->cur_playing
            && gtk_tree_row_reference_valid(mwin->plist->cur_playing))
    {
        path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
        do {
            indices = gtk_tree_path_get_indices(path);
            if(indices && indices[0] == 0
                    && xfmedia_settings_get_bool("/xfmedia/playlist/repeat"))
            {
                tot_items = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL);
                g_free(path);
                path = gtk_tree_path_new_from_indices(tot_items-1, -1);
            } else if(!indices || indices[0] != 0)
                gtk_tree_path_prev(path);
            else
                break;
        } while(!xfmedia_playlist_play_file_at_path(mwin, path));
        g_free(path);
    }
}

void
xfmedia_next_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")
            && mwin->plist->cur_playing)
    {
        DBG("putting item on shuffle history");
        g_queue_push_head(mwin->plist->shuffle_history,
                gtk_tree_row_reference_copy(mwin->plist->cur_playing));
    }
    
    xfmedia_playlist_play_file_at_path_persistent(mwin, NULL);
}

gboolean
xfmedia_pos_lbl_click_cb(GtkWidget *w, GdkEventButton *evt, gpointer user_data)
{
    if(evt->button == 1) {
        xfmedia_settings_set_bool("/xfmedia/general/show_remaining",
                !xfmedia_settings_get_bool("/xfmedia/general/show_remaining"));
        xfmedia_position_cb(user_data);
    }
    
    return xfmedia_mainwin_menu_cb(w, evt, user_data);
}

void
xfmedia_shuffle_menu_toggle_cb(GtkCheckMenuItem *mi, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean set;
    
    set = gtk_check_menu_item_get_active(mi);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mwin->plist->shuffle_btn), set);
    xfmedia_settings_set_bool("/xfmedia/playlist/shuffle", set);
    
    if(!set && !g_queue_is_empty(mwin->plist->shuffle_history)) {
        gint i, tot = g_queue_get_length(mwin->plist->shuffle_history);
        for(i = tot-1; i >= 0; i--)
            gtk_tree_row_reference_free((GtkTreeRowReference *)g_queue_pop_nth(mwin->plist->shuffle_history, i));
    }
    
    if(set)
        xfmedia_playlist_shuffle_unset_all(mwin->plist);
}

void
xfmedia_repeat_menu_toggle_cb(GtkCheckMenuItem *mi, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean set;
    
    set = gtk_check_menu_item_get_active(mi);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mwin->plist->repeat_btn), set);
    xfmedia_settings_set_bool("/xfmedia/playlist/repeat", set);
}

void
xfmedia_repeat_one_menu_toggle_cb(GtkCheckMenuItem *mi, gpointer user_data)
{
    gboolean set;
    
    set = gtk_check_menu_item_get_active(mi);
    xfmedia_settings_set_bool("/xfmedia/playlist/repeat_one", set);
}

void
xfmedia_show_vis_menu_toggle_cb(GtkCheckMenuItem *mi, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean set, has_video;

    set = gtk_check_menu_item_get_active(mi);
    xfmedia_settings_set_bool("/xfmedia/general/show_audio_vis", set);
    
    has_video = gtk_xine_get_stream_info(mwin->gtx, XINE_STREAM_INFO_HAS_VIDEO);
    if(!set) {
        if(!has_video)
            gtk_widget_hide(mwin->video_window);
        gtk_xine_set_vis(mwin->gtx, NULL);
    } else {
        if(!has_video && gtk_xine_get_status(mwin->gtx) == XINE_STATUS_PLAY) {
            gint width, height;
            GdkGeometry geom;
            
            gtk_widget_show(mwin->video_window);
            gtk_xine_set_vis(mwin->gtx,
                    xfmedia_settings_get_string("/xfmedia/general/selected_vis"));
            
            width = gtk_xine_get_stream_info(mwin->gtx,
                    XINE_STREAM_INFO_VIDEO_WIDTH);
            height = gtk_xine_get_stream_info(mwin->gtx,
                    XINE_STREAM_INFO_VIDEO_HEIGHT);
            if(width < 1 || height < 1) {
                width = 480;
                height = 360;
            }
            gtk_window_resize(GTK_WINDOW(mwin->video_window),
                    width, height);
            geom.min_width = geom.min_height = 1;
            geom.min_aspect = geom.max_aspect = (gdouble)width / height;
            DBG("window size: %dx%d, aspect: %f", width, height, geom.max_aspect);
            gtk_window_set_geometry_hints(GTK_WINDOW(mwin->video_window),
                    GTK_WIDGET(mwin->gtx), &geom,
                    GDK_HINT_MIN_SIZE|GDK_HINT_ASPECT);
        }
    }
}

gboolean
xfmedia_vwin_delete_cb(GtkWidget *w, GdkEvent *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    DBG("entering");
    
    if(gtk_xine_get_stream_info(mwin->gtx, XINE_STREAM_INFO_HAS_VIDEO)) {
        gtk_xine_stop(mwin->gtx);
        DBG("checking plist (%p)", mwin->plist);
        xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_EXT);
    } else {
        gtk_xine_set_vis(mwin->gtx, NULL);
        gtk_widget_hide(w);
        gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->svis_mi), FALSE);
        xfmedia_settings_set_bool("/xfmedia/general/show_audio_vis", FALSE);
    }
    
    /* never let the window get destroyed here */
    return TRUE;
}

gboolean
xfmedia_autohidewin_enter_notify(GtkWidget *w, GdkEventCrossing *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint x, y;
    
    x = xfmedia_settings_get_int("/xfmedia/general/pos_x");
    y = xfmedia_settings_get_int("/xfmedia/general/pos_y");
    
    gtk_window_move(GTK_WINDOW(mwin->window), x, y);
    if(xfmedia_settings_get_bool("/xfmedia/general/window_sticky"))
        gtk_window_stick(GTK_WINDOW(mwin->window));
    else
        gtk_window_unstick(GTK_WINDOW(mwin->window));
    if(xfmedia_settings_get_bool("/xfmedia/general/window_above"))
        gtk_window_set_keep_above(GTK_WINDOW(mwin->window), TRUE);
    gtk_widget_show(mwin->window);
    gtk_widget_hide(mwin->autohide_window);
    
    return FALSE;
}

gboolean
xfmedia_mainwin_enterleave_notify_cb(GtkWidget *w, GdkEventCrossing *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(evt->type == GDK_ENTER_NOTIFY) {
        if(mwin->autohide_timer) {
            g_source_remove(mwin->autohide_timer);
            mwin->autohide_timer = 0;
        }
    } else if(evt->type == GDK_LEAVE_NOTIFY) {
        if(evt->detail == GDK_NOTIFY_NONLINEAR_VIRTUAL
                && !mwin->temp_disable_autohide
                && xfmedia_settings_get_bool("/xfmedia/general/dock_autohide")
                && !mwin->autohide_timer)
        {
            mwin->autohide_timer = g_timeout_add(AUTOHIDE_TIMEOUT,
                    (GSourceFunc)xfmedia_mainwin_do_autohide, mwin);
        }
    }
    
    return FALSE;
}

GdkFilterReturn
xfmedia_mainwin_decs_filter(GdkXEvent *gxevt, GdkEvent *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    XCrossingEvent *xevt = (XCrossingEvent *)gxevt;

    if(xevt->type == EnterNotify) {
        if(mwin->autohide_timer) {
            g_source_remove(mwin->autohide_timer);
            mwin->autohide_timer = 0;
        }
    } else if(xevt->type == LeaveNotify) {
        if(xevt->detail == NotifyNonlinearVirtual
                && !(xevt->state & (ButtonPressMask|ButtonMotionMask
                                    |Button1MotionMask|Button2MotionMask
                                    |Button3MotionMask|Button4MotionMask
                                    |Button5MotionMask))
                && !mwin->temp_disable_autohide
                && xfmedia_settings_get_bool("/xfmedia/general/dock_autohide")
                && !mwin->autohide_timer)
        {
            mwin->autohide_timer = g_timeout_add(AUTOHIDE_TIMEOUT,
                    (GSourceFunc)xfmedia_mainwin_do_autohide, mwin);
        }
    } else if(xevt->type == MotionNotify) {
        if(mwin->autohide_timer) {
            g_source_remove(mwin->autohide_timer);
            mwin->autohide_timer = 0;
        }
    }
    
    return GDK_FILTER_CONTINUE;
}

gboolean
xfmedia_mainwin_configure_cb(GtkWidget *w, GdkEventConfigure *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint x = -1, y = -1;
    
    if(GTK_WIDGET_VISIBLE(w))
        gtk_window_get_position(GTK_WINDOW(w), &x, &y);
    else
        return FALSE;
    
    xfmedia_settings_set_int("/xfmedia/general/pos_x", x);
    xfmedia_settings_set_int("/xfmedia/general/pos_y", y);

    xfmedia_settings_set_int("/xfmedia/general/width", evt->width);
    if(gtk_expander_get_expanded(GTK_EXPANDER(mwin->playlist_pane)))
        xfmedia_settings_set_int("/xfmedia/general/height_large", evt->height);
    else
        xfmedia_settings_set_int("/xfmedia/general/height_small", evt->height);
    
    return FALSE;
}

gboolean
xfmedia_mainwin_winstate_cb(GtkWidget *w, GdkEventWindowState *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    mwin->is_iconified = (evt->new_window_state & GDK_WINDOW_STATE_ICONIFIED);
    
    if(xfmedia_settings_get_bool("/xfmedia/general/minimize_to_tray")
            && mwin->is_iconified)
    {
        /*DBG("setting skip taskbar hint");*/
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(w), TRUE);
        gtk_widget_hide(mwin->window);
        gtk_window_deiconify(GTK_WINDOW(mwin->window));
    } else {
        /*DBG("unsetting skip taskbar hint");*/
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(w), FALSE);
    }
    
    return FALSE;
}

/* NOTE: this is a dirty hack and needs to be fixed */
static gboolean
expand_treescroll_idle(gpointer data)
{
    XfMediaMainwin *mwin = data;
    
    gdk_threads_enter();
    xfmedia_playlist_treeview_scroll_cb(gtk_tree_view_get_vadjustment(GTK_TREE_VIEW(mwin->plist->treeview)), mwin);
    gdk_threads_leave();
    
    return FALSE;
}

void
xfmedia_mainwin_expander_activate_cb(GtkExpander *widget, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint h_l, h_s, w;
    
    if(gtk_expander_get_expanded(widget)) {
        gtk_window_get_size(GTK_WINDOW(mwin->window), &w, &h_s);
        h_l = xfmedia_settings_get_int("/xfmedia/general/height_large");
        if(h_l == -1) {
            h_l = h_s * 3;
            xfmedia_settings_set_int("/xfmedia/general/height_large", h_l);
        }
        xfmedia_settings_set_int("/xfmedia/general/height_small", h_s);
        xfmedia_settings_set_int("/xfmedia/general/width", w);
        
        gtk_widget_set_size_request(mwin->window, -1, -1);
        gtk_window_resize(GTK_WINDOW(mwin->window), w, h_l);
        gtk_widget_show(mwin->plist->playlist_widget);
        xfmedia_settings_set_bool("/xfmedia/playlist/playlist_expanded", TRUE);
        g_timeout_add(350, expand_treescroll_idle, mwin);
    } else {
        gtk_window_get_size(GTK_WINDOW(mwin->window), &w, &h_l);
        gtk_widget_hide(mwin->plist->playlist_widget);
        h_s = xfmedia_settings_get_int("/xfmedia/general/height_small");
        gtk_widget_set_size_request(mwin->window, -1, h_s);
        gtk_window_resize(GTK_WINDOW(mwin->window), w, h_s);
        xfmedia_settings_set_int("/xfmedia/general/height_large", h_l);
        xfmedia_settings_set_int("/xfmedia/general/width", w);
        xfmedia_settings_set_bool("/xfmedia/playlist/playlist_expanded", FALSE);
    }
}

void
xfmedia_about_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    XfceAboutInfo *ainfo;
    GtkWidget *dlg, *toplevel = NULL;
    
    if(w)
        toplevel = gtk_widget_get_toplevel(w);
    if(!toplevel)
        toplevel = mwin->window;
    
    ainfo = xfce_about_info_new("Xfmedia", VERSION,
            _("Xfmedia is a lightweight media player,\nbased on the xine engine"),
            XFCE_COPYRIGHT_TEXT("2004", "Brian Tarricone"), XFCE_LICENSE_GPL);
    xfce_about_info_set_homepage(ainfo, "http://spuriousinterrupt.org/projects/xfmedia/");
    xfce_about_info_add_credit(ainfo, "Brian Tarricone", "bjt23@cornell.edu",
            "Original Author");
    
    dlg = xfce_about_dialog_new(GTK_WINDOW(toplevel), ainfo, NULL);
    g_signal_connect(G_OBJECT(dlg), "response",
            G_CALLBACK(gtk_widget_destroy), NULL);
    gtk_widget_show_all(dlg);
}

gboolean
xfmedia_vwin_button_press_cb(GtkWidget *w, GdkEventButton *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint x, y;
    
    if(evt->button == 1 && evt->type == GDK_2BUTTON_PRESS) {
        gboolean is_fs = gtk_xine_is_fullscreen(mwin->gtx);
        
        gtk_xine_set_fullscreen(mwin->gtx, !is_fs);
        if(!is_fs) {
            gtk_window_get_position(GTK_WINDOW(mwin->video_window), &x, &y);
            xfmedia_settings_set_int("/xfmedia/general/vwin_pos_x", x);
            xfmedia_settings_set_int("/xfmedia/general/vwin_pos_y", y);
            gtk_widget_hide(mwin->video_window);
        } else {
            x = xfmedia_settings_get_int("/xfmedia/general/vwin_pos_x");
            y = xfmedia_settings_get_int("/xfmedia/general/vwin_pos_y");
            gtk_window_move(GTK_WINDOW(mwin->video_window), x, y);
            gtk_widget_show(mwin->video_window);
        }
        
        return TRUE;
    }
    
    return FALSE;
}

gboolean
xfmedia_vwin_motion_notify_cb(GtkWidget *w, GdkEventMotion *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GdkScreen *gscreen;
    gint sheight, swidth;
    
    if(!gtk_xine_is_fullscreen(mwin->gtx))
        return FALSE;
    
    gscreen = gtk_widget_get_screen(mwin->window);
    swidth = gdk_screen_get_width(gscreen);
    sheight = gdk_screen_get_height(gscreen);
    
    if(!GTK_WIDGET_VISIBLE(mwin->fs_controls) && evt->y == sheight - 1) {
        Window fs_window;
        
        fs_window = GDK_WINDOW_XWINDOW(gtk_xine_get_fullscreen_window(mwin->gtx));
        gdk_property_change(mwin->fs_controls->window,
                gdk_atom_intern("WM_TRANSIENT_FOR", FALSE),
                gdk_atom_intern("WINDOW", FALSE),
                32, GDK_PROP_MODE_REPLACE,
                (const guchar *)&fs_window, sizeof(Window));
        
        gtk_widget_set_size_request(mwin->fs_controls, swidth, -1);
        gtk_widget_realize(mwin->fs_controls);
        gtk_window_move(GTK_WINDOW(mwin->fs_controls), 0,
                sheight - mwin->fs_controls->allocation.height);
        gtk_widget_show(mwin->fs_controls);
    } else if(GTK_WIDGET_VISIBLE(mwin->fs_controls)
            && evt->y < sheight - mwin->fs_controls->allocation.height)
    {
        gtk_widget_hide(mwin->fs_controls);
    }
    
    return FALSE;
}

gboolean
xfmedia_vwin_mapunmap_cb(GtkWidget *w, GdkEvent *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    int visible;
    
    if(evt->type == GDK_MAP)
        visible = 1;
    else if(evt->type == GDK_UNMAP)
        visible = 0;
    else
        return FALSE;
    
    gtk_xine_port_send_gui_data(mwin->gtx, XINE_GUI_SEND_VIDEOWIN_VISIBLE,
            &visible);
    
    return FALSE;
}

void
xfmedia_gtkxine_stream_ended_cb(GtkXine *gtx, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean ret = FALSE;
    
    if(!xfmedia_settings_get_bool("/xfmedia/playlist/repeat_one")
            && xfmedia_settings_get_bool("/xfmedia/playlist/shuffle"))
    {
        DBG("putting item on shuffle history");
        g_queue_push_head(mwin->plist->shuffle_history,
                gtk_tree_row_reference_copy(mwin->plist->cur_playing));
    }
    
    xfmedia_mainwin_set_play_buttons(mwin);
    xfmedia_mainwin_disconnect_slider_callback(mwin);
    gtk_range_set_value(GTK_RANGE(mwin->position_slider), 0.0);
    gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), 0.0);
    xfmedia_mainwin_connect_slider_callback(mwin);
    gtk_label_set_markup(GTK_LABEL(mwin->time_label),
            "<span font_family=\"monospace\" size=\"large\" weight=\"bold\"> 00:00</span>");
    gtk_label_set_markup(GTK_LABEL(mwin->fs_time_label),
            "<span font_family=\"monospace\"> 00:00</span>");
    
    xfmedia_plugins_forward_signal_VOID("stream-ended");
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/auto_advance")) {
        if(xfmedia_settings_get_bool("/xfmedia/playlist/repeat_one")) {
            GtkTreePath *path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
            ret = xfmedia_playlist_play_file_at_path(mwin, path);
            gtk_tree_path_free(path);
        }
        if(!ret)
            ret = xfmedia_playlist_play_file_at_path_persistent(mwin, NULL);
    }
    
    if(!ret && GTK_WIDGET_VISIBLE(mwin->video_window))
        gtk_widget_hide(mwin->video_window);
}

void
xfmedia_gtkxine_ui_message_cb(GtkXine *gtx, const gchar *message,
        gpointer user_data)
{
    DBG("got UI message: '%s'", message);
}

void
xfmedia_gtkxine_progress_message_cb(GtkXine *gtx, const gchar *message, gint percent,
        gpointer user_data)
{
    DBG("got progress message: '%s' at %d%%", message, percent);
}

void
xfmedia_gtkxine_format_changed_cb(GtkXine *gtx, gint width, gint height, gint aspect_code,
        gboolean is_pan_scan, gpointer user_data)
{
    DBG("got format changed signal: %dx%d, aspect_code: %d, is_pan_scan: %s",
            width, height, aspect_code, is_pan_scan?"yes":"no");
}

void
xfmedia_mainwin_quit_mi_cb(GtkMenuItem *mi, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_EXT);
}

gboolean
xfmedia_mainwin_map_cb(GtkWidget *w, GdkEvent *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GdkWindow *decs;
    
    decs = xfmedia_get_window_decorations(w->window);
    
    gdk_error_trap_push();
    
    if(decs) {
        mwin->window_decs = decs;
        gdk_error_trap_push();
        gdk_window_set_events(decs, gdk_window_get_events(decs)
                | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
                | GDK_POINTER_MOTION_MASK);
        gdk_window_add_filter(decs, xfmedia_mainwin_decs_filter, mwin);
    }
    
    gdk_error_trap_pop();
    
    return FALSE;
}

gboolean
xfmedia_mainwin_unmap_cb(GtkWidget *w, GdkEvent *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(mwin->window_decs) {
        gdk_error_trap_push();
        gdk_window_remove_filter(mwin->window_decs,
                xfmedia_mainwin_decs_filter, mwin);
        g_object_unref(G_OBJECT(mwin->window_decs));
        mwin->window_decs = NULL;
        gdk_error_trap_pop();
    }
    
    return FALSE;
}

gboolean
xfmedia_mainwin_keypress_cb(GtkWidget *w, GdkEventKey *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    const gchar *keybind_id;
    
    if(GTK_WIDGET_HAS_FOCUS(mwin->plist->treeview)
            && (evt->keyval == GDK_Up || evt->keyval == GDK_KP_Up
                || evt->keyval == GDK_Down || evt->keyval == GDK_KP_Down
                || evt->keyval == GDK_Return || evt->keyval == GDK_KP_Enter))
    {
        /* eat these, as they should be for playlist navigation */
        DBG("  eaten");
        return FALSE;
    }
    
    keybind_id = xfmedia_keybindings_lookup(evt->keyval, evt->state);
    if(keybind_id)
        xfmedia_keybindings_activate(mwin, keybind_id);
    
    return FALSE;
}

static void
mainwin_popupmenu_deactivate_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    guint sigid = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(w),
            "xfmedia-deactivate-sigid"));
    
    g_signal_handler_disconnect(G_OBJECT(w), sigid);
    mwin->temp_disable_autohide--;
}

gboolean
xfmedia_mainwin_menu_cb(GtkWidget *w, GdkEventButton *evt, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(evt->button == 3 && mwin->popup_menu) {
        guint sigid = g_signal_connect(G_OBJECT(mwin->popup_menu), "deactivate",
                G_CALLBACK(mainwin_popupmenu_deactivate_cb), mwin);
        g_object_set_data(G_OBJECT(mwin->popup_menu),
                "xfmedia-deactivate-sigid", GUINT_TO_POINTER(sigid));
        mwin->temp_disable_autohide++;
        gtk_menu_popup(GTK_MENU(mwin->popup_menu), NULL, NULL, NULL, NULL,
                evt->button, evt->time);
        
        return TRUE;
    }
    
    return FALSE;
}

static GtkWidget *
tray_icon_menu_create(XfMediaMainwin *mwin)
{
    GtkWidget *menu, *mi;
    gint iw, ih;
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &iw, &ih);
    
    menu = gtk_menu_new();
    gtk_widget_show(menu);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("P_revious Track"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_prev_cb), mwin);
    
    mwin->tray_icon_playpause_btn = mi =
            gtk_image_menu_item_new_with_mnemonic(_("_Play"));
    mwin->tray_icon_play_img = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY,
            GTK_ICON_SIZE_MENU);
    g_object_ref(G_OBJECT(mwin->tray_icon_play_img));
    if(gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED) != XINE_SPEED_NORMAL
            || gtk_xine_get_status(mwin->gtx) != XINE_STATUS_PLAY)
    {
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
                mwin->tray_icon_play_img);
    }
    mwin->tray_icon_pause_img = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PAUSE,
            GTK_ICON_SIZE_MENU);
    g_object_ref(G_OBJECT(mwin->tray_icon_pause_img));
    if(gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED) == XINE_SPEED_NORMAL
            && gtk_xine_get_status(mwin->gtx) == XINE_STATUS_PLAY)
    {
        gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(mi))),
                _("_Pause"));
        gtk_label_set_use_underline(GTK_LABEL(gtk_bin_get_child(GTK_BIN(mi))),
                TRUE);
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
                mwin->tray_icon_pause_img);
    }
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playpause_cb), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Stop"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_stop_cb), mwin);
    
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Next Track"));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi),
            gtk_image_new_from_stock(GTK_STOCK_MEDIA_NEXT, GTK_ICON_SIZE_MENU));
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_next_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    mi = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect_swapped(G_OBJECT(mi), "activate",
            G_CALLBACK(gtk_widget_destroy), mwin->window);
    
    return menu;
}

gboolean
xfmedia_tray_icon_button_press_cb(GtkWidget *w, GdkEventButton *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if(evt->button == 1) {
        if(GTK_WIDGET_VISIBLE(mwin->window))
            gtk_widget_hide(mwin->window);
        else if(GTK_WIDGET_VISIBLE(mwin->autohide_window))
            gtk_widget_hide(mwin->autohide_window);
        else {
            gtk_window_move(GTK_WINDOW(mwin->window),
                    xfmedia_settings_get_int("/xfmedia/general/pos_x"),
                    xfmedia_settings_get_int("/xfmedia/general/pos_y"));
            if(xfmedia_settings_get_bool("/xfmedia/general/window_sticky"))
                gtk_window_stick(GTK_WINDOW(mwin->window));
            else
                gtk_window_unstick(GTK_WINDOW(mwin->window));
            if(xfmedia_settings_get_bool("/xfmedia/general/window_above"))
                gtk_window_set_keep_above(GTK_WINDOW(mwin->window), TRUE);
            gtk_widget_show(mwin->window);
        }
        return TRUE;
    } else if(evt->button == 3) {
        if(!mwin->tray_icon_menu)
            mwin->tray_icon_menu = tray_icon_menu_create(mwin);
        gtk_menu_popup(GTK_MENU(mwin->tray_icon_menu), NULL, NULL, NULL,
                NULL, evt->button, evt->time);
        return TRUE;
    }
    
    return FALSE;
}

void
xfmedia_mainwin_mediamarks_changed_cb(gpointer data)
{
    XfMediaMainwin *mwin = data;
    GtkWidget *old_menu, *menu;
    
    old_menu = gtk_menu_item_get_submenu(GTK_MENU_ITEM(mwin->mmarks_mi));
    gtk_widget_destroy(old_menu);
    
    menu = xfmedia_mediamarks_create_menu(mwin);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(mwin->mmarks_mi), menu);
}
