/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#include <gdk/gdkkeysyms.h>

#include <gtk/gtk.h>
#if !GTK_CHECK_VERSION(2, 4, 0)
#define GTK_STOCK_NETWORK GTK_STOCK_OPEN
#define GTK_STOCK_HARDDISK GTK_STOCK_OPEN
#define g_queue_get_length(x) (x->length)
#endif

#include <libxfcegui4/libxfcegui4.h>

#include "exo/exo-ellipsized-label.h"

#include "playlist.h"
#include "mainwin.h"
#include <xfmedia/xfmedia-settings.h>
#include "xfmedia-internal.h"
#include "mediamarks.h"

#ifndef XINE_META_INFO_TRACK_NUMBER
#define XINE_META_INFO_TRACK_NUMBER 11
#endif

static void
treeview_popupmenu_deactivate_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    guint sigid = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(w),
            "xfmedia-deactivate-sigid"));
    
    g_signal_handler_disconnect(G_OBJECT(w), sigid);
    mwin->temp_disable_autohide--;
}

gboolean
xfmedia_playlist_treeview_keypress_cb(GtkWidget *w, GdkEventKey *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    if((evt->keyval == GDK_Return || evt->keyval == GDK_KP_Enter) ) {
        GList *selected;
        
        selected = gtk_tree_selection_get_selected_rows(mwin->plist->tree_sel, NULL);
        if(selected) {
            if(!g_queue_is_empty(mwin->plist->shuffle_history)) {
                gint i, tot = g_queue_get_length(mwin->plist->shuffle_history);
                for(i = tot-1; i >= 0; i--)
                    gtk_tree_row_reference_free((GtkTreeRowReference *)g_queue_pop_nth(mwin->plist->shuffle_history, i));
            }
            
            xfmedia_playlist_shuffle_unset_all(mwin->plist);
            
            xfmedia_playlist_play_file_at_path(mwin, (GtkTreePath *)selected->data);
            
            g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
            g_list_free(selected);
        }
        
        return TRUE;
    } else if(evt->keyval == GDK_Delete || evt->keyval == GDK_KP_Delete) {
        xfmedia_playlist_remove_cb(w, mwin->plist);
        return TRUE;
    }
    
    return FALSE;
}

gboolean
xfmedia_playlist_treeview_buttonpress_cb(GtkWidget *w, GdkEventButton *evt,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GList *selected;
    
    if(evt->button == 1 && evt->type == GDK_2BUTTON_PRESS) {
        selected = gtk_tree_selection_get_selected_rows(mwin->plist->tree_sel, NULL);
        if(selected) {
            if(!g_queue_is_empty(mwin->plist->shuffle_history)) {
                gint i, tot = g_queue_get_length(mwin->plist->shuffle_history);
                for(i = tot-1; i >= 0; i--)
                    gtk_tree_row_reference_free((GtkTreeRowReference *)g_queue_pop_nth(mwin->plist->shuffle_history, i));
            }
            
            xfmedia_playlist_shuffle_unset_all(mwin->plist);
            
            xfmedia_playlist_play_file_at_path_persistent(mwin,
                    (GtkTreePath *)selected->data);
            
            /* free selection stuff */
            g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
            g_list_free(selected);
        }
    } else if(evt->button == 3) {
        GtkTreePath *mouse_over = NULL;
        guint sigid;
        
        if(gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(mwin->plist->treeview),
                evt->x, evt->y, &mouse_over, NULL, NULL, NULL))
        {
            gtk_tree_selection_unselect_all(mwin->plist->tree_sel);
            gtk_tree_selection_select_path(mwin->plist->tree_sel, mouse_over);
            gtk_tree_path_free(mouse_over);
        }
        
        sigid = g_signal_connect(G_OBJECT(mwin->plist->popup_menu), "deactivate",
                G_CALLBACK(treeview_popupmenu_deactivate_cb), mwin);
        g_object_set_data(G_OBJECT(mwin->plist->popup_menu),
                "xfmedia-deactivate-sigid", GUINT_TO_POINTER(sigid));
        mwin->temp_disable_autohide++;
        gtk_menu_popup(GTK_MENU(mwin->plist->popup_menu), NULL, NULL, NULL,
                NULL, evt->button, evt->time);
        return TRUE;  /* eat event so DnD doesn't kick in */
    }
    
    return FALSE;
}

void
xfmedia_playlist_add_url_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *dlg, *lbl, *entry, *topvbox, *hbox;
    
    mwin->temp_disable_autohide++;
    
    dlg = gtk_dialog_new_with_buttons(_("Add URL"), GTK_WINDOW(mwin->window),
            GTK_DIALOG_NO_SEPARATOR|GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OK,
            GTK_RESPONSE_ACCEPT, NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_ACCEPT);
    gtk_container_set_border_width(GTK_CONTAINER(dlg), BORDER);
    topvbox = GTK_DIALOG(dlg)->vbox;
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(topvbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new_with_mnemonic(_("_URL:"));
    gtk_widget_show(lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), entry);
    
    gtk_widget_set_size_request(dlg, 400, -1);
    if(gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_ACCEPT) {
        gchar *url = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
        
        xfmedia_mainwin_add_file(mwin, url, -1, TRUE);
        xfmedia_playlist_shuffle_add_entries(mwin->plist, -1, 1);
        xfmedia_plugins_forward_signal_STRING("playlist-entry-added", url);
        mwin->plist->is_dirty = TRUE;
        mwin->plist->save_on_exit = TRUE;
        g_free(url);
    }
    
    gtk_widget_destroy(dlg);
    
    mwin->temp_disable_autohide--;
}

static guint
recurse_add_dirs(XfMediaMainwin *mwin, const gchar *dirname)
{
    GDir *dir;
    const gchar *file;
    gchar fullpath[PATH_MAX];
    GList *file_list = NULL, *dir_list = NULL, *l;
    guint nadded = 0;
    
    dir = g_dir_open(dirname, 0, NULL);
    if(!dir)
        return 0;
    
    while((file = g_dir_read_name(dir))) {
        g_snprintf(fullpath, PATH_MAX, "%s%s%s", dirname, G_DIR_SEPARATOR_S, file);
        if(g_file_test(fullpath, G_FILE_TEST_IS_DIR)) {
            dir_list = g_list_insert_sorted(dir_list, g_strdup(fullpath),
                    (GCompareFunc)g_ascii_strcasecmp);
        } else {
            file_list = g_list_insert_sorted(file_list, g_strdup(fullpath),
                    (GCompareFunc)g_ascii_strcasecmp);
        }
    }
    g_dir_close(dir);
    
    for(l = dir_list; l; l = l->next) {
        recurse_add_dirs(mwin, (const gchar *)l->data);
        g_free(l->data);
    }
    if(dir_list)
        g_list_free(dir_list);
    
    gdk_threads_enter();
    for(l = file_list; l; l = l->next) {
        xfmedia_mainwin_add_file(mwin, (const gchar *)l->data, -1, TRUE);
        nadded++;
        xfmedia_plugins_forward_signal_STRING("playlist-entry-added",
                (const gchar *)l->data);
        g_free(l->data);
    }
    gdk_threads_leave();
    
    if(file_list)
        g_list_free(file_list);
    
    return nadded;
}

struct RecurseData {
    XfMediaMainwin *mwin;
    gchar *basedir;
};

static gpointer
recurse_add_dirs_th(gpointer data)
{
    struct RecurseData *recurse_data = data;
    guint nadded;
    
    nadded = recurse_add_dirs(recurse_data->mwin, recurse_data->basedir);
    
    if(nadded > 0) {
        gdk_threads_enter();
        xfmedia_playlist_shuffle_add_entries(recurse_data->mwin->plist,
                -1, nadded);
        gdk_threads_leave();
    }
    
    g_free(recurse_data);
    
    return NULL;
}

void
xfmedia_playlist_add_dir_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *chooser;
    struct RecurseData *recurse_data;
#define MY_FOLDER_SELECT_RESPONSE 749837
    
    mwin->temp_disable_autohide++;
    
    chooser = xfce_file_chooser_new(_("Add Directory"), GTK_WINDOW(mwin->window),
            XFCE_FILE_CHOOSER_ACTION_SELECT_FOLDER, GTK_STOCK_CANCEL,
            GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN,
            MY_FOLDER_SELECT_RESPONSE, NULL);
    if(mwin->plist->chooser_last_dir)
        xfce_file_chooser_set_current_folder(XFCE_FILE_CHOOSER(chooser),
                mwin->plist->chooser_last_dir);
    gtk_dialog_set_default_response(GTK_DIALOG(chooser), GTK_RESPONSE_ACCEPT);
    
    gtk_widget_show(chooser);
    if(gtk_dialog_run(GTK_DIALOG(chooser)) == MY_FOLDER_SELECT_RESPONSE) {
        gchar *dirname;
        
        dirname = xfce_file_chooser_get_filename(XFCE_FILE_CHOOSER(chooser));
        if(dirname) {
            recurse_data = g_new0(struct RecurseData, 1);
            recurse_data->mwin = mwin;
            recurse_data->basedir = dirname;
            mwin->plist->is_dirty = TRUE;
            mwin->plist->save_on_exit = TRUE;
            g_thread_create((GThreadFunc)recurse_add_dirs_th, recurse_data,
                    FALSE, NULL);
        }
    }
    
    gtk_widget_destroy(chooser);
    
    mwin->temp_disable_autohide--;
#undef MY_FOLDER_SELECT_RESPONSE
}

void
xfmedia_playlist_add_file_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *chooser, *btn;
    XfceFileFilter *filter;
    GSList *l, *filenames;
    gboolean defer = TRUE;
    gint resp;
    guint nadded = 0;
    
    mwin->temp_disable_autohide++;
    
    chooser = xfce_file_chooser_new(_("Add Media File"), GTK_WINDOW(mwin->window),
            XFCE_FILE_CHOOSER_ACTION_OPEN, NULL);
    btn = xfmedia_custom_button_new(_("Add _Selected"), GTK_STOCK_ADD);
    gtk_widget_show(btn);
    gtk_dialog_add_action_widget(GTK_DIALOG(chooser), btn, GTK_RESPONSE_APPLY);
    gtk_dialog_add_button(GTK_DIALOG(chooser), GTK_STOCK_CLOSE, GTK_RESPONSE_CANCEL);
    btn = xfmedia_custom_button_new(_("A_dd and Close"), GTK_STOCK_ADD);
    GTK_WIDGET_SET_FLAGS(btn, GTK_CAN_DEFAULT);
    gtk_widget_show(btn);
    gtk_dialog_add_action_widget(GTK_DIALOG(chooser), btn, GTK_RESPONSE_ACCEPT);
    xfce_file_chooser_set_select_multiple(XFCE_FILE_CHOOSER(chooser), TRUE);
    if(mwin->plist->chooser_last_dir)
        xfce_file_chooser_set_current_folder(XFCE_FILE_CHOOSER(chooser),
                mwin->plist->chooser_last_dir);
    gtk_dialog_set_default_response(GTK_DIALOG(chooser), GTK_RESPONSE_ACCEPT);
    
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("All Files"));
    xfce_file_filter_add_pattern(filter, "*");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("Music Files"));
    xfce_file_filter_add_pattern(filter, "*.aif*");
    xfce_file_filter_add_pattern(filter, "*.au");
    xfce_file_filter_add_pattern(filter, "*.mka");
    xfce_file_filter_add_pattern(filter, "*.mpa");
    xfce_file_filter_add_pattern(filter, "*.mp2");
    xfce_file_filter_add_pattern(filter, "*.mp3");
    xfce_file_filter_add_pattern(filter, "*.ogg");
    xfce_file_filter_add_pattern(filter, "*.wav");
    xfce_file_filter_add_pattern(filter, "*.wma");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("Playlists"));
    xfce_file_filter_add_pattern(filter, "*.m3u");
    xfce_file_filter_add_pattern(filter, "*.pls");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("Video Files"));
    xfce_file_filter_add_pattern(filter, "*.avi");
    xfce_file_filter_add_pattern(filter, "*.mkv");
    xfce_file_filter_add_pattern(filter, "*.mpeg");
    xfce_file_filter_add_pattern(filter, "*.mpg");
    xfce_file_filter_add_pattern(filter, "*.ogg");
    xfce_file_filter_add_pattern(filter, "*.ogm");
    xfce_file_filter_add_pattern(filter, "*.vob");
    xfce_file_filter_add_pattern(filter, "*.wmv");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    
    gtk_widget_show(chooser);
    do {
        resp = gtk_dialog_run(GTK_DIALOG(chooser));
        if(resp == GTK_RESPONSE_ACCEPT)
            gtk_widget_hide(chooser);
        if(resp == GTK_RESPONSE_ACCEPT || resp == GTK_RESPONSE_APPLY) {
            filenames = xfce_file_chooser_get_filenames(XFCE_FILE_CHOOSER(chooser));
            if(filenames && g_slist_length(filenames) < 10)
                defer = FALSE;
            for(l = filenames; l; l = l->next) {
                if(xfmedia_playlist_is_playlist_file(l->data))
                    xfmedia_playlist_load(l->data, mwin, FALSE);
                else {
                    xfmedia_mainwin_add_file(mwin, l->data, -1, defer);
                    nadded++;
                }
                xfmedia_plugins_forward_signal_STRING("playlist-entry-added",
                        (const gchar *)l->data);
                g_free(l->data);
            }
            if(filenames)
                g_slist_free(filenames);
            if(mwin->plist->chooser_last_dir)
                g_free(mwin->plist->chooser_last_dir);
            mwin->plist->chooser_last_dir = xfce_file_chooser_get_current_folder(XFCE_FILE_CHOOSER(chooser));
            mwin->plist->is_dirty = TRUE;
            mwin->plist->save_on_exit = TRUE;
            
            xfmedia_playlist_shuffle_add_entries(mwin->plist, -1, nadded);
        }
    } while(resp == GTK_RESPONSE_APPLY);
        
    gtk_widget_destroy(chooser);
    
    mwin->temp_disable_autohide--;
}

static void
add_menu_position(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, 
        gpointer user_data)
{
    GtkRequisition req;
    gint xpointer, ypointer;
    
    gtk_widget_size_request(GTK_WIDGET(menu), &req);
    
    /* get pointer position */
    gdk_display_get_pointer(
            gdk_screen_get_display(gtk_widget_get_screen(GTK_WIDGET(menu))),
            NULL, &xpointer, &ypointer, NULL);
    
    /* set x to the current position */
    *x = xpointer;
    /* ...and y to the current position minus the menu height */
    *y = ypointer - req.height;
    
    /* FIXME: huh? */
    *push_in = FALSE;
}

static void
add_menu_deactivated_cb(GtkMenu *menu, gpointer user_data)
{
    guint sigid = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(menu),
            "xfmedia-deactivate-sigid"));
    XfMediaMainwin *mwin = g_object_get_data(G_OBJECT(menu), "xfmedia-mwin");
    GtkToggleButton *tb = user_data;
    
    gtk_toggle_button_set_active(tb, FALSE);
    g_signal_handler_disconnect(G_OBJECT(menu), sigid);
    
    /* restore menu */
    mwin->temp_disable_autohide--;
}

static void
playlist_add_autoplay_cb(GtkMenuItem *mi, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *lbl = gtk_bin_get_child(GTK_BIN(mi));
    const gchar *plugin_id = gtk_label_get_text(GTK_LABEL(lbl));
    GList *mrls, *l;
    guint nadded = 0;
    
    if(plugin_id) {
        mrls = gtk_xine_get_autoplay_mrls(mwin->gtx, plugin_id);
        for(l = mrls; l; l = l->next) {
            xfmedia_mainwin_add_file(mwin, (const gchar *)l->data, -1, TRUE);
            nadded++;
            xfmedia_plugins_forward_signal_STRING("playlist-entry-added",
                    (const gchar *)l->data);
        }
        if(mrls) {
            g_list_free(mrls);
            mwin->plist->save_on_exit = TRUE;
            xfmedia_playlist_shuffle_add_entries(mwin->plist, -1, nadded);
        }
    }
}

void
xfmedia_playlist_add_cb(GtkToggleButton *tb, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    static GtkWidget *menu = NULL;
    guint sigid;
    
    if(!gtk_toggle_button_get_active(tb))
        return;
    
    if(!menu) {
        GtkWidget *mi, *img, *ap_menu;
        GList *autoplay_plugins, *l;
        
        menu = gtk_menu_new();
        gtk_widget_show(menu);
        
        /* xine's autoplay menu */
        autoplay_plugins = gtk_xine_get_autoplay_input_plugin_ids(mwin->gtx);
        if(autoplay_plugins) {
            img = gtk_image_new_from_stock(GTK_STOCK_NEW, GTK_ICON_SIZE_MENU);
            gtk_widget_show(img);
            mi = gtk_image_menu_item_new_with_mnemonic(_("_Other"));
            gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
            gtk_widget_show(mi);
            gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
            
            ap_menu = gtk_menu_new();
            gtk_widget_show(ap_menu);
            gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), ap_menu);
            
            for(l = autoplay_plugins; l; l = l->next) {
                mi = gtk_image_menu_item_new_with_label((const gchar *)l->data);
                gtk_widget_show(mi);
                if(strstr((const gchar *)l->data, "DVD") == l->data
                        || strstr((const gchar *)l->data, "CD") == l->data
                        || strstr((const gchar *)l->data, "VCD") == l->data)
                {
                    img = gtk_image_new_from_stock(GTK_STOCK_CDROM,
                            GTK_ICON_SIZE_MENU);
                    gtk_widget_show(img);
                    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
                }
                gtk_menu_shell_append(GTK_MENU_SHELL(ap_menu), mi);
                g_signal_connect(G_OBJECT(mi), "activate",
                        G_CALLBACK(playlist_add_autoplay_cb), mwin);
            }
            g_list_free(autoplay_plugins);
        }
        
        img = gtk_image_new_from_stock(GTK_STOCK_NETWORK, GTK_ICON_SIZE_MENU);
        gtk_widget_show(img);
        mi = gtk_image_menu_item_new_with_mnemonic(_("_URL..."));
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
        g_signal_connect(G_OBJECT(mi), "activate",
                G_CALLBACK(xfmedia_playlist_add_url_cb), mwin);
        
        img = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_MENU);
        gtk_widget_show(img);
        mi = gtk_image_menu_item_new_with_mnemonic(_("_Directory..."));
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
        g_signal_connect(G_OBJECT(mi), "activate",
                G_CALLBACK(xfmedia_playlist_add_dir_cb), mwin);
        
        img = gtk_image_new_from_stock(GTK_STOCK_HARDDISK, GTK_ICON_SIZE_MENU);
        gtk_widget_show(img);
        mi = gtk_image_menu_item_new_with_mnemonic(_("_File..."));
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
        gtk_widget_show(mi);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
        g_signal_connect(G_OBJECT(mi), "activate",
                G_CALLBACK(xfmedia_playlist_add_file_cb), mwin);
                
        g_object_set_data(G_OBJECT(menu), "xfmedia-mwin", mwin);
    }
    
    /* temporarily disable autohide, so the menu popup doesn't make the window
     * hide itself */
    mwin->temp_disable_autohide++;
    
    sigid = g_signal_connect(G_OBJECT(menu), "deactivate",
            G_CALLBACK(add_menu_deactivated_cb), tb);
    g_object_set_data(G_OBJECT(menu), "xfmedia-deactivate-sigid",
            GUINT_TO_POINTER(sigid));
    
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL,
            (GtkMenuPositionFunc)add_menu_position, tb, 0,
            gtk_get_current_event_time());
}

void
xfmedia_playlist_remove_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaPlaylist *plist = user_data;
    GList *selected, *l, *rrefs = NULL;
    guint low_index = -1, index;
    
    selected = gtk_tree_selection_get_selected_rows(plist->tree_sel, NULL);
    if(selected) {
        for(l = selected; l; l = l->next) {
            GtkTreePath *path = l->data;
            
            index = xfmedia_tree_path_to_index(path);
            if(index < low_index)
                low_index = index;
            
            rrefs = g_list_prepend(rrefs,
                gtk_tree_row_reference_new(GTK_TREE_MODEL(plist->file_list), path));
            gtk_tree_path_free(path);
        }
        g_list_free(selected);
        
        for(l = rrefs; l; l = l->next) {
            GtkTreeRowReference *rref = l->data;
            GtkTreePath *path = gtk_tree_row_reference_get_path(rref);
            gint path_index = xfmedia_tree_path_to_index(path);
            
            if(xfmedia_playlist_remove_file(plist, path, -1))
                xfmedia_playlist_shuffle_remove_entries(plist, path_index, 1);
            gtk_tree_path_free(path);
            gtk_tree_row_reference_free(rref);
        }
        g_list_free(rrefs);
    }
    
    DBG("renumbering from %d", low_index);
    if(low_index != -1)
        xfmedia_playlist_renumber_file_list(plist->file_list, low_index, -1);        
}

void
xfmedia_playlist_go_up_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaPlaylist *plist = user_data;
    GtkTreeIter me, above_me;
    gint index, path_index;
    GList *selected;
    
    selected = gtk_tree_selection_get_selected_rows(plist->tree_sel, NULL);
    if(selected && g_list_length(selected) == 1) {
        if(gtk_tree_model_get_iter(GTK_TREE_MODEL(plist->file_list), &me,
                (GtkTreePath *)selected->data))
        {
            if(gtk_tree_path_prev((GtkTreePath *)selected->data)) {
                if(gtk_tree_model_get_iter(GTK_TREE_MODEL(plist->file_list),
                        &above_me, (GtkTreePath *)selected->data))
                {
                    gtk_tree_model_get(GTK_TREE_MODEL(plist->file_list),
                            &above_me, PLAYLIST_COL_NUM, &index, -1);
                    gtk_list_store_set(plist->file_list, &above_me,
                            PLAYLIST_COL_NUM, index+1, -1);
                    gtk_list_store_set(plist->file_list, &me,
                            PLAYLIST_COL_NUM, index, -1);
                    gtk_list_store_swap(plist->file_list, &me, &above_me);
                    path_index = xfmedia_tree_path_to_index((GtkTreePath *)selected->data);
                    xfmedia_playlist_shuffle_swap_entries(plist, path_index, path_index-1);
                    plist->is_dirty = TRUE;
                    plist->save_on_exit = TRUE;
                }
            }
        }
    }
    
    if(selected) {
        g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
        g_list_free(selected);
    }
}

void
xfmedia_playlist_go_down_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaPlaylist *plist = user_data;
    GtkTreeIter me, below_me;
    gint index, path_index;
    GList *selected;
    
    selected = gtk_tree_selection_get_selected_rows(plist->tree_sel, NULL);
    if(selected && g_list_length(selected) == 1) {
        if(gtk_tree_model_get_iter(GTK_TREE_MODEL(plist->file_list), &me,
                (GtkTreePath *)selected->data))
        {
            gtk_tree_path_next((GtkTreePath *)selected->data);
            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(plist->file_list),
                    &below_me, (GtkTreePath *)selected->data))
            {
                gtk_tree_model_get(GTK_TREE_MODEL(plist->file_list),
                        &below_me, PLAYLIST_COL_NUM, &index, -1);
                gtk_list_store_set(plist->file_list, &below_me,
                        PLAYLIST_COL_NUM, index-1, -1);
                gtk_list_store_set(plist->file_list, &me,
                        PLAYLIST_COL_NUM, index, -1);
                gtk_list_store_swap(plist->file_list, &me, &below_me);
                path_index = xfmedia_tree_path_to_index((GtkTreePath *)selected->data);
                xfmedia_playlist_shuffle_swap_entries(plist, path_index, path_index+1);
                plist->is_dirty = TRUE;
                plist->save_on_exit = TRUE;
            }
        }
    }
    
    if(selected) {
        g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
        g_list_free(selected);
    }
}

static gint
playlist_dirty_confirm(XfMediaMainwin *mwin, GtkWindow *parent)
{
    GtkWidget *msgdlg, *btn;
    gint resp;
    gchar message[2048];
    const gchar *primary = _("Save changes to this playlist before closing?");
    const gchar *secondary = _("You've changed this playlist since it was last saved.");
    
    mwin->temp_disable_autohide++;
    
#if GTK_CHECK_VERSION(2, 4, 0)
    g_snprintf(message, 2048,
            "<span size=\"larger\" weight=\"bold\">%s</span>\n\n%s",
            primary, secondary);
    msgdlg = gtk_message_dialog_new_with_markup(parent,
            GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE, message);
#else
    g_snprintf(message, 2048, "%s\n\n%s", primary, secondary);
    msgdlg = gtk_message_dialog_new(parent,
            GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE, message);
#endif
    btn = xfmedia_custom_button_new(_("Do_n't save"), GTK_STOCK_NO);
    gtk_widget_show(btn);
    gtk_dialog_add_action_widget(GTK_DIALOG(msgdlg), btn, GTK_RESPONSE_NO);
    gtk_dialog_add_buttons(GTK_DIALOG(msgdlg), GTK_STOCK_CANCEL,
            GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_YES, NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(msgdlg), GTK_RESPONSE_YES);
    resp = gtk_dialog_run(GTK_DIALOG(msgdlg));
    gtk_widget_destroy(msgdlg);
    
    mwin->temp_disable_autohide--;
    
    return resp;
}

void
xfmedia_playlist_new_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gint resp;
    
    if(mwin->plist->is_dirty) {
        resp = playlist_dirty_confirm(mwin, GTK_WINDOW(mwin->window));
        switch(resp) {
            case GTK_RESPONSE_CANCEL:
                return;
            case GTK_RESPONSE_YES:
                if(!xfmedia_playlist_save_as_cb(mwin->time_label, mwin))
                    return;
                break;
            case GTK_RESPONSE_NO:
                break;
        }
    }
    
    gtk_list_store_clear(mwin->plist->file_list);
    mwin->plist->is_dirty = FALSE;
    mwin->plist->save_on_exit = TRUE;
    xfmedia_settings_set_int("/xfmedia/playlist/playlist_position", -1);
    xfmedia_playlist_shuffle_init_entries(mwin->plist, 0);
    
    xfmedia_plugins_forward_signal_VOID("playlist-cleared");
}

gboolean
xfmedia_playlist_save_as_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *topwin, *chooser;
    gchar *filename;
    gboolean ret = FALSE;
    
    if(!gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL))
        return TRUE;
    
    mwin->temp_disable_autohide++;
    
    topwin = gtk_widget_get_toplevel(w);
    
    chooser = xfce_file_chooser_new(_("Save Playlist"), GTK_WINDOW(topwin),
            XFCE_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL,
            GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
    xfce_file_chooser_set_select_multiple(XFCE_FILE_CHOOSER(chooser), FALSE);
    if(mwin->plist->chooser_last_dir)
        xfce_file_chooser_set_current_folder(XFCE_FILE_CHOOSER(chooser),
                mwin->plist->chooser_last_dir);
    gtk_dialog_set_default_response(GTK_DIALOG(chooser), GTK_RESPONSE_ACCEPT);
    gtk_widget_show(chooser);
    
    if(gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
        gtk_widget_hide(chooser);
        filename = xfce_file_chooser_get_filename(XFCE_FILE_CHOOSER(chooser));
        if(filename) {
            ret = xfmedia_playlist_save(mwin->plist->file_list, filename);
            xfmedia_plugins_forward_signal_STRING("playlist-saved", filename);
            g_free(filename);
        }
    }
    gtk_widget_destroy(chooser);
    
    mwin->temp_disable_autohide--;
    
    return ret;
}

void
xfmedia_playlist_open_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkWidget *topwin, *chooser;
    XfceFileFilter *filter;
    gchar *filename;
    gint resp;
    
    mwin->temp_disable_autohide++;
    
    topwin = gtk_widget_get_toplevel(w);
    
    if(mwin->plist->is_dirty) {
        resp = playlist_dirty_confirm(mwin, GTK_WINDOW(mwin->window));
        switch(resp) {
            case GTK_RESPONSE_CANCEL:
                mwin->temp_disable_autohide--;
                return;
            case GTK_RESPONSE_YES:
                if(!xfmedia_playlist_save_as_cb(mwin->time_label, mwin)) {
                    mwin->temp_disable_autohide--;
                    return;
                }
                break;
            case GTK_RESPONSE_NO:
                break;
        }
    }
    
    chooser = xfce_file_chooser_new(_("Open Playlist"), GTK_WINDOW(topwin),
            XFCE_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL,
            GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
    xfce_file_chooser_set_select_multiple(XFCE_FILE_CHOOSER(chooser), FALSE);
    if(mwin->plist->chooser_last_dir)
        xfce_file_chooser_set_current_folder(XFCE_FILE_CHOOSER(chooser),
                mwin->plist->chooser_last_dir);
    gtk_dialog_set_default_response(GTK_DIALOG(chooser), GTK_RESPONSE_ACCEPT);
    
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("All Files"));
    xfce_file_filter_add_pattern(filter, "*");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    filter = xfce_file_filter_new();
    xfce_file_filter_set_name(filter, _("Playlists"));
    xfce_file_filter_add_pattern(filter, "*.m3u");
    xfce_file_filter_add_pattern(filter, "*.pls");
    xfce_file_chooser_add_filter(XFCE_FILE_CHOOSER(chooser), filter);
    
    gtk_widget_show(chooser);
    
    if(gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
        gtk_widget_hide(chooser);
        filename = xfce_file_chooser_get_filename(XFCE_FILE_CHOOSER(chooser));
        if(filename) {
            if(xfmedia_playlist_load(filename, mwin, TRUE)) {
                mwin->plist->is_dirty = FALSE;
                mwin->plist->save_on_exit = TRUE;
                xfmedia_settings_set_int("/xfmedia/playlist/playlist_position", -1);
                xfmedia_plugins_forward_signal_STRING("playlist-loaded", filename);
            }
            g_free(filename);
        }
        if(mwin->plist->chooser_last_dir)
            g_free(mwin->plist->chooser_last_dir);
        mwin->plist->chooser_last_dir = xfce_file_chooser_get_current_folder(XFCE_FILE_CHOOSER(chooser));
    }
    gtk_widget_destroy(chooser);
    
    mwin->temp_disable_autohide--;
}

static gboolean
xfmedia_playlist_get_visible_range(XfMediaMainwin *mwin, GtkTreePath **start,
        GtkTreePath **end)
{
    GtkTreeView *treeview = GTK_TREE_VIEW(mwin->plist->treeview);
    GdkRectangle rect;
    gint cell_x = 0, cell_y = 0;
    gint wx = 0, wy = 0;
    GtkTreeIter itr;
    gboolean path_good = FALSE, path_end_good = FALSE;
    
    g_return_val_if_fail(mwin && start && end, FALSE);
    
    /* get visible rectangle */
    gtk_tree_view_get_visible_rect(treeview, &rect);
    
    /* get a tree path and iter for the bottommost visible item */
    gtk_tree_view_tree_to_widget_coords(treeview, rect.x,
            rect.y+rect.height-1, &wx, &wy);
    if(gtk_tree_view_get_path_at_pos(treeview, wx, wy, end, NULL,
            &cell_x, &cell_y))
    {
        if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                &itr, *end))
        {
            path_end_good = TRUE;
        }
    }
    
    /* get a tree path and iter for the topmost visible item */
    gtk_tree_view_tree_to_widget_coords(treeview, rect.x, rect.y, &wx, &wy);
    if(gtk_tree_view_get_path_at_pos(treeview, wx, wy, start, NULL,
            &cell_x, &cell_y))
    {
        if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                &itr, *start))
        {
            path_good = TRUE;
        }
    }
    
    if(path_good && path_end_good)
        return TRUE;
    else
        return FALSE;
}

static gboolean
xfmedia_playlist_load_info_idled(gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GtkTreePath *path_start, *path_end, *path;
    GtkTreeIter itr;
    gboolean info_loaded = FALSE;
    gchar *name = NULL, *filename = NULL;
    gint song_len = 0;
    
    gdk_threads_enter();
    
    if(!xfmedia_playlist_get_visible_range(mwin, &path_start, &path_end)) {
        gdk_threads_leave();
        return FALSE;
    }
    
    /* move path_end one farther */
    gtk_tree_path_next(path_end);
    
    for(path = path_start;
        gtk_tree_path_compare(path, path_end);
        gtk_tree_path_next(path))
    {
        if(path && gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                &itr, path))
        {
            gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list), &itr,
                    PLAYLIST_COL_FILENAME, &filename,
                    PLAYLIST_COL_INFOLOADED, &info_loaded, -1);
            if(filename && !info_loaded) {
                if(xfmedia_mainwin_get_file_info(mwin, filename, &name,
                        &song_len))
                {
                    if(!name)
                        name = xfmedia_filename_to_name(filename);
                    xfmedia_playlist_set_info(mwin->plist->file_list,
                            &itr, name, song_len);
                    g_free(name);
                }
                g_free(filename);
            }
        }
        
        while(gtk_events_pending())
            gtk_main_iteration();
    }
    
    gtk_tree_path_free(path_start);
    gtk_tree_path_free(path_end);
    
    gdk_threads_leave();
    
    return FALSE;
}

void
xfmedia_playlist_treeview_scroll_cb(GtkAdjustment *adj, gpointer user_data)
{
    if(xfmedia_settings_get_int("/xfmedia/playlist/metadata_load_impetus") == 0)
        g_idle_add(xfmedia_playlist_load_info_idled, user_data);
}

void
xfmedia_playlist_shuffle_toggle_cb(GtkToggleButton *b, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean set;
    
    set = gtk_toggle_button_get_active(b);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->shuffle_mi), set);
    xfmedia_settings_set_bool("/xfmedia/playlist/shuffle", set);
    
    if(!set && !g_queue_is_empty(mwin->plist->shuffle_history)) {
        gint i, tot = g_queue_get_length(mwin->plist->shuffle_history);
        for(i = tot-1; i >= 0; i--)
            gtk_tree_row_reference_free((GtkTreeRowReference *)g_queue_pop_nth(mwin->plist->shuffle_history, i));
    }
    
    if(set)
        xfmedia_playlist_shuffle_unset_all(mwin->plist);
}

void
xfmedia_playlist_repeat_toggle_cb(GtkToggleButton *b, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gboolean set;
    
    set = gtk_toggle_button_get_active(b);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mwin->repeat_mi), set);
    xfmedia_settings_set_bool("/xfmedia/playlist/repeat", set);
}

void
xfmedia_playlist_file_info_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GList *selected;
#define TMPBUF 64
    gchar *filename = NULL, tmpbuf[TMPBUF], *tmpbuf1 = NULL;
    const gchar *value;
    GtkTreePath *path;
    GtkTreeIter itr;
    GtkWidget *dlg, *lbl, *entry, *topvbox, *hbox, *topwin, *frame, *vbox;
    GtkSizeGroup *sg;
    struct stat st;
    gint pos_stream, pos_length, length_time = 0;
    guint32 fourcc;
    
    selected = gtk_tree_selection_get_selected_rows(mwin->plist->tree_sel, NULL);
    if(!selected)
        return;
    path = g_list_nth_data(selected, 0);
    if(selected->next)
        g_list_foreach(selected->next, (GFunc)gtk_tree_path_free, NULL);
    g_list_free(selected);
    
    if(!gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list), &itr, path)) {
        gtk_tree_path_free(path);
        return;
    }
    gtk_tree_path_free(path);
    
    gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list), &itr,
            PLAYLIST_COL_FILENAME, &filename, -1);
    if(!filename)
        return;
    
    if(!xine_open(mwin->extra_stream, filename)) {
        xfce_message_dialog(GTK_WINDOW(mwin->window), "Xfmedia",
                GTK_STOCK_DIALOG_INFO, _("Unable to open file."),
                _("Xfmedia was unable to open the selected file.  Be sure that the file exists and that you have permission to access it."),
                GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
        g_free(filename);
        return;
    }
    
    topwin = gtk_widget_get_toplevel(mwin->plist->treeview);
    dlg = gtk_dialog_new_with_buttons(_("File Info"), GTK_WINDOW(topwin),
            GTK_DIALOG_NO_SEPARATOR|GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_STOCK_CLOSE, GTK_RESPONSE_NONE, NULL);
    gtk_container_set_border_width(GTK_CONTAINER(dlg), BORDER);
    topvbox = GTK_DIALOG(dlg)->vbox;
    gtk_box_set_spacing(GTK_BOX(topvbox), BORDER/2);
    
    sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    
    frame = xfce_framebox_new(_("Filename"), TRUE);
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(topvbox), frame, FALSE, FALSE, 0);
    
    lbl = exo_ellipsized_label_new(filename);
    exo_ellipsized_label_set_ellipsize(EXO_ELLIPSIZED_LABEL(lbl),
            EXO_PANGO_ELLIPSIZE_START);
    gtk_label_set_selectable(GTK_LABEL(lbl), TRUE);
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_label_set_text(GTK_LABEL(lbl), filename);
    gtk_widget_show(lbl);
    xfce_framebox_add(XFCE_FRAMEBOX(frame), lbl);
    
    frame = xfce_framebox_new(_("Media Info"), TRUE);
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(topvbox), frame, FALSE, FALSE, 0);
    
    vbox = gtk_vbox_new(FALSE, BORDER);
    gtk_widget_show(vbox);
    xfce_framebox_add(XFCE_FRAMEBOX(frame), vbox);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Title:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_TITLE);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Artist:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_ARTIST);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Album:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_ALBUM);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Genre:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_GENRE);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    lbl = gtk_label_new(_("Track:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_TRACK_NUMBER);
    if(value) {
        gtk_entry_set_text(GTK_ENTRY(entry), value);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(value));
    } else
        gtk_entry_set_width_chars(GTK_ENTRY(entry), 3);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Year:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_YEAR);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_entry_set_width_chars(GTK_ENTRY(entry), 4);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Comment:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    value = xine_get_meta_info(mwin->extra_stream, XINE_META_INFO_COMMENT);
    if(value)
        gtk_entry_set_text(GTK_ENTRY(entry), value);
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    hbox = gtk_hbox_new(FALSE, BORDER);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    
    lbl = gtk_label_new(_("Length:"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_size_group_add_widget(sg, lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    if(xine_get_pos_length(mwin->extra_stream, &pos_stream, &pos_length, &length_time)) {
        length_time /= 1000;
        g_snprintf(tmpbuf, TMPBUF, "%02d:%02d", length_time/60, length_time%60);
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf));
    }
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    lbl = gtk_label_new(_("File size (b):"));
    gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
    gtk_widget_show(lbl);
    gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
    
    entry = gtk_entry_new();
    if(!stat(filename, &st)) {
        g_snprintf(tmpbuf, TMPBUF, "%d", (gint)st.st_size);
        tmpbuf1 = xfmedia_num_prettify(tmpbuf);
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf1);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf1));
        g_free(tmpbuf1);
    }
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_widget_show(entry);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    
    if(xine_get_stream_info(mwin->extra_stream, XINE_STREAM_INFO_HAS_AUDIO)) {
        frame = xfce_framebox_new(_("Audio Info"), TRUE);
        gtk_widget_show(frame);
        gtk_box_pack_start(GTK_BOX(topvbox), frame, FALSE, FALSE, 0);
        
        vbox = gtk_vbox_new(FALSE, BORDER);
        gtk_widget_show(vbox);
        xfce_framebox_add(XFCE_FRAMEBOX(frame), vbox);
        
        hbox = gtk_hbox_new(FALSE, BORDER);
        gtk_widget_show(hbox);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
        
        lbl = gtk_label_new(_("Codec:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_size_group_add_widget(sg, lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        value = xine_get_meta_info(mwin->extra_stream,
                XINE_META_INFO_AUDIOCODEC);
        if(value)
            gtk_entry_set_text(GTK_ENTRY(entry), value);
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("Samples/sec:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_AUDIO_SAMPLERATE));
        tmpbuf1 = xfmedia_num_prettify(tmpbuf);
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf1);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf1));
        g_free(tmpbuf1);
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        hbox = gtk_hbox_new(FALSE, BORDER);
        gtk_widget_show(hbox);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
        
        lbl = gtk_label_new(_("Bits/sample:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_size_group_add_widget(sg, lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_AUDIO_BITS));
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("Channels:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_AUDIO_CHANNELS));
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("Bitrate (kbit):"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_AUDIO_BITRATE)/1000);
        tmpbuf1 = xfmedia_num_prettify(tmpbuf);
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf1);
        g_free(tmpbuf1);
        gtk_entry_set_width_chars(GTK_ENTRY(entry),
                strlen(gtk_entry_get_text(GTK_ENTRY(entry))));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    }
    
    if(xine_get_stream_info(mwin->extra_stream, XINE_STREAM_INFO_HAS_VIDEO)) {
        frame = xfce_framebox_new(_("Video Info"), TRUE);
        gtk_widget_show(frame);
        gtk_box_pack_start(GTK_BOX(topvbox), frame, FALSE, FALSE, 0);
        
        vbox = gtk_vbox_new(FALSE, BORDER);
        gtk_widget_show(vbox);
        xfce_framebox_add(XFCE_FRAMEBOX(frame), vbox);
        
        hbox = gtk_hbox_new(FALSE, BORDER);
        gtk_widget_show(hbox);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
        
        lbl = gtk_label_new(_("Codec:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_size_group_add_widget(sg, lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        value = xine_get_meta_info(mwin->extra_stream,
                XINE_META_INFO_VIDEOCODEC);
        if(value)
            gtk_entry_set_text(GTK_ENTRY(entry), value);
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("FourCC:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        fourcc = xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_VIDEO_FOURCC);
        tmpbuf[0] = ((gchar *)&fourcc)[0];
        tmpbuf[1] = ((gchar *)&fourcc)[1];
        tmpbuf[2] = ((gchar *)&fourcc)[2];
        tmpbuf[3] = ((gchar *)&fourcc)[3];
        tmpbuf[4] = 0;
        if(g_utf8_validate(tmpbuf, -1, NULL))
            gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf1);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), 4);
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        hbox = gtk_hbox_new(FALSE, BORDER);
        gtk_widget_show(hbox);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
        
        lbl = gtk_label_new(_("Frame size:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_size_group_add_widget(sg, lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%dx%d",
                xine_get_stream_info(mwin->extra_stream,
                    XINE_STREAM_INFO_VIDEO_WIDTH),
                xine_get_stream_info(mwin->extra_stream,
                    XINE_STREAM_INFO_VIDEO_HEIGHT));
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("Channels:"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_VIDEO_CHANNELS));
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf);
        gtk_entry_set_width_chars(GTK_ENTRY(entry), strlen(tmpbuf));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        
        lbl = gtk_label_new(_("Bitrate (kbit):"));
        gtk_misc_set_alignment(GTK_MISC(lbl), 0.0, 0.5);
        gtk_widget_show(lbl);
        gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
        
        entry = gtk_entry_new();
        g_snprintf(tmpbuf, TMPBUF, "%d", xine_get_stream_info(mwin->extra_stream,
                XINE_STREAM_INFO_VIDEO_BITRATE)/1000);
        tmpbuf1 = xfmedia_num_prettify(tmpbuf);
        gtk_entry_set_text(GTK_ENTRY(entry), tmpbuf1);
        g_free(tmpbuf1);
        gtk_entry_set_width_chars(GTK_ENTRY(entry),
                strlen(gtk_entry_get_text(GTK_ENTRY(entry))));
        gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
        gtk_widget_show(entry);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    }
    
    xine_close(mwin->extra_stream);
    
    gtk_dialog_run(GTK_DIALOG(dlg));
    gtk_widget_destroy(dlg);

    g_free(filename);
#undef TMPBUF
}

void
xfmedia_playlist_treeview_drag_get_cb(GtkWidget *widget,
        GdkDragContext *drag_context, GtkSelectionData *data, guint info,
        guint time, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GdkAtom playlist_item = gdk_atom_intern("PLAYLIST_ITEM", FALSE);
    
    if(data->target == playlist_item) {
        GtkTreeRowReference *srcrref = g_object_get_data(G_OBJECT(drag_context),
                "gtk-tree-view-source-row");
        GtkTreePath *sourcerow = gtk_tree_row_reference_get_path(srcrref);
        GtkTreeIter *itr;
        
        if(!sourcerow)
            return;
        
        itr  = g_new0(GtkTreeIter, 1);
        gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                itr, sourcerow);
        
        gtk_selection_data_set(data, playlist_item, 8,
                (gpointer)&itr, sizeof(itr));
        
        gtk_tree_path_free(sourcerow);
    }
}

void
xfmedia_playlist_treeview_drag_rcv_cb(GtkWidget *widget,
        GdkDragContext *drag_context, gint x, gint y,  GtkSelectionData *data,
        guint info, guint time, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GdkAtom _PLAYLIST_ITEM = gdk_atom_intern("PLAYLIST_ITEM", FALSE);
    GdkAtom _URI_LIST = gdk_atom_intern("text/uri-list", FALSE);
    GdkAtom _STRING = gdk_atom_intern("STRING", FALSE);
    GtkTreePath *dest_path = NULL;
    GtkTreeViewDropPosition drop_pos;
    GtkTreeIter *src_itr, dest_itr;
    gboolean drag_success = FALSE;
    
    DBG("got DnD target: %s", gdk_atom_name(data->target));
    
    if(data->target == _PLAYLIST_ITEM) {
        src_itr = *(GtkTreeIter **)data->data;
        if(!src_itr) {
            gtk_drag_finish(drag_context, FALSE, FALSE, time);
            return;
        }
        if(gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(mwin->plist->treeview),
                    x, y, &dest_path, &drop_pos))
        {
            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                                       &dest_itr, dest_path)
                    && gtk_list_store_iter_is_valid(mwin->plist->file_list,
                                                    src_itr))
            {
                gint src_index = 0, dest_index = 0;
                
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                        src_itr, PLAYLIST_COL_NUM, &src_index, -1);
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                        &dest_itr, PLAYLIST_COL_NUM, &dest_index, -1);
                
				if(src_index != dest_index) {
					src_index--;
					dest_index--;
					
					DBG("src_index=%d, dest_index=%d", src_index, dest_index);
					
					if(drop_pos == GTK_TREE_VIEW_DROP_BEFORE
							|| drop_pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
					{
						DBG("putting src before dest");
						/* gtk oddity: if dest is the first item in the list, then
						 * gtk_list_store_move_before(ls, src, dest) will move src
						 * to the _end_ of the list.  somehow it thinks the item
						 * before the zeroth item is the last item. this sucks. */
						if(dest_index == 0) {
							GtkTreeModel *model = GTK_TREE_MODEL(mwin->plist->file_list);
							/* first convert our iters into row refs */
							GtkTreePath *srcpath = gtk_tree_model_get_path(model, src_itr);
							GtkTreePath *destpath = gtk_tree_model_get_path(model, &dest_itr);
							GtkTreeRowReference *srcref =
								gtk_tree_row_reference_new(model, srcpath);
							GtkTreeRowReference *destref =
								gtk_tree_row_reference_new(model, destpath);
							gtk_tree_path_free(srcpath);
							gtk_tree_path_free(destpath);
							
							/* then move the src _after_ the dest */
							gtk_list_store_move_after(mwin->plist->file_list,
									src_itr, &dest_itr);
							
							/* now get new iters from the row refs */
							srcpath = gtk_tree_row_reference_get_path(srcref);
							destpath = gtk_tree_row_reference_get_path(destref);
							gtk_tree_model_get_iter(model, src_itr, srcpath);
							gtk_tree_model_get_iter(model, &dest_itr, destpath);
							gtk_tree_path_free(srcpath);
							gtk_tree_path_free(destpath);
							
							/* swap the two items, so now dest is before src */
							gtk_list_store_swap(mwin->plist->file_list, src_itr,
									&dest_itr);
							
							/* ditch the row refs */
							gtk_tree_row_reference_free(srcref);
							gtk_tree_row_reference_free(destref);
							
							/* man, that was stupid */
						} else {
							gtk_list_store_move_before(mwin->plist->file_list,
									src_itr, &dest_itr);
						}
					} else {
						DBG("putting src after dest");
						gtk_list_store_move_after(mwin->plist->file_list, src_itr,
								&dest_itr);
					}
					
					if(src_index < dest_index)
						xfmedia_playlist_renumber_file_list(mwin->plist->file_list,
								src_index, dest_index);
					else
						xfmedia_playlist_renumber_file_list(mwin->plist->file_list,
								dest_index, src_index);
					xfmedia_playlist_shuffle_swap_entries(mwin->plist,
							src_index, dest_index);
					
					mwin->plist->is_dirty = TRUE;
					drag_success = TRUE;
				}
            } else {
                DBG("one of the itrs is invalid");
            }
            
            gtk_tree_path_free(dest_path);
        } else {
            DBG("gtk_tree_view_get_dest_row_at_pos() failed");
        }
        
        g_free(src_itr);
    } else if(data->target == _URI_LIST) {
        GList *uris, *l;
        gint from_index = -1, index, num_added = 0;
        
        DBG("we have an uri list: '%s'", (gchar *)data->data);
        
        if(gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(mwin->plist->treeview),
                    x, y, &dest_path, &drop_pos))
        {
            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                    &dest_itr, dest_path))
            {
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                        &dest_itr, PLAYLIST_COL_NUM, &from_index, -1);
                if(drop_pos == GTK_TREE_VIEW_DROP_BEFORE
                        || drop_pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                {
                    from_index--;
                }
            }
            gtk_tree_path_free(dest_path);
        }
        
        index = from_index;
        uris = xfmedia_uri_list_extract_uris((const gchar *)data->data);
        for(l = uris; l; l = l->next) {
            gchar *filename = l->data;
            
            if(!g_ascii_strncasecmp(filename, "file://", 7))
                filename += 7;
            DBG("adding '%s'", filename);
            xfmedia_mainwin_add_file(mwin, filename, index, TRUE);
            g_free(l->data);
            if(index != -1)
                index++;
            num_added++;
        }
        if(uris) {
            if(from_index == -1) {
                from_index = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL);
                from_index -= num_added - 1;
            }
            xfmedia_playlist_renumber_file_list(mwin->plist->file_list,
                    from_index, -1);
            g_list_free(uris);
            mwin->plist->save_on_exit = TRUE;
            
            xfmedia_playlist_shuffle_add_entries(mwin->plist, from_index, num_added);
        }
        
        drag_success = TRUE;
    } else if(data->target == _STRING) {
        gchar *filename = data->data;
        gint from_index = -1;
        
        if(gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(mwin->plist->treeview),
                    x, y, &dest_path, &drop_pos))
        {
            if(gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                    &dest_itr, dest_path))
            {
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                        &dest_itr, PLAYLIST_COL_NUM, &from_index, -1);
                if(drop_pos == GTK_TREE_VIEW_DROP_BEFORE
                        || drop_pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                {
                    from_index--;
                }
            }
            gtk_tree_path_free(dest_path);
        }
        
        if(filename) {
            if(!g_ascii_strncasecmp(filename, "file://", 7))
                filename += 7;
            xfmedia_mainwin_add_file(mwin, filename, from_index, TRUE);
            xfmedia_playlist_renumber_file_list(mwin->plist->file_list,
                    from_index, -1);
            xfmedia_playlist_shuffle_add_entries(mwin->plist, from_index, 1);
            mwin->plist->save_on_exit = TRUE;
            drag_success = TRUE;
        }
    }
    
    gtk_drag_finish(drag_context, drag_success, FALSE, time);
}

void
xfmedia_playlist_add_mm_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    GList *selected;
    
    selected = gtk_tree_selection_get_selected_rows(mwin->plist->tree_sel, NULL);
    if(selected) {
        xfmedia_mediamarks_add_path(mwin, GTK_WINDOW(mwin->window),
                (GtkTreePath *)selected->data);
        
        g_list_foreach(selected, (GFunc)gtk_tree_path_free, NULL);
        g_list_free(selected);
    }
}
