/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#include <gtk/gtk.h>
#if !GTK_CHECK_VERSION(2, 4, 0)
#define gtk_event_box_set_visible_window(x, y)
#define g_queue_get_length(x) (x->length)
#endif

#include <libxfcegui4/libxfcegui4.h>

#include "exo/exo-ellipsized-label.h"
#include "exo/exo-cell-renderer-ellipsized-text.h"

#include "playlist.h"
#include "mainwin.h"
#include <xfmedia/xfmedia-settings.h>
#include "xfmedia-internal.h"

gboolean
xfmedia_playlist_play_file_at_path(XfMediaMainwin *mwin, GtkTreePath *path)
{
    gchar *filename = NULL, *name = NULL;
    gint song_len = 0, row_num = 0, width, height;
    GtkTreeIter itr;
    gboolean ret = FALSE, has_video, was_queued = FALSE;
    GdkRectangle tree_visible;
    GtkTreePath *path_top = NULL, *path_bottom = NULL;
    GtkTreeView *treeview = GTK_TREE_VIEW(mwin->plist->treeview);
    
    mwin->plist->cur_playing_len = -1;
    
    if(!gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list), &itr, path))
        return FALSE;
    
    row_num = xfmedia_tree_path_to_index(path);
    gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list), &itr,
            PLAYLIST_COL_FILENAME, &filename,
            PLAYLIST_COL_QUEUED, &was_queued,
            -1);
    if(!filename)
        return FALSE;
    
    /* figure out if the playlist entry at |path| is visible, and if not,
       scroll to it. */
    /* top item */
    gtk_tree_view_get_path_at_pos(treeview, 0, 0, &path_top, NULL, NULL, NULL);
    /* bottom item */
    gtk_tree_view_get_visible_rect(treeview, &tree_visible);
    gtk_tree_view_get_path_at_pos(treeview, 0, tree_visible.height-2,
            &path_bottom, NULL, NULL, NULL);
    if(path_top && path_bottom) {
        /* and compare */
        gint path_index = xfmedia_tree_path_to_index(path);
        if(xfmedia_tree_path_to_index(path_top) > path_index
                || xfmedia_tree_path_to_index(path_bottom) < path_index)
        {
            gtk_tree_view_scroll_to_cell(treeview, path, NULL, TRUE, 0.5, 0.0);
        }
    } else
        gtk_tree_view_scroll_to_cell(treeview, path, NULL, TRUE, 0.5, 0.0);
    
    if(path_top)
        gtk_tree_path_free(path_top);
    if(path_bottom)
        gtk_tree_path_free(path_bottom);
    
    if(mwin->plist->cur_playing)
        gtk_tree_row_reference_free(mwin->plist->cur_playing);
    mwin->plist->cur_playing = gtk_tree_row_reference_new(GTK_TREE_MODEL(mwin->plist->file_list), path);
    gtk_widget_queue_draw(mwin->plist->treeview);
    xfmedia_settings_set_int("/xfmedia/playlist/playlist_position", row_num);
    
    DBG("about to call gtk_xine_open()");
    
    if(gtk_xine_open(mwin->gtx, filename)) {
        gboolean was_visible = (GTK_WIDGET_VISIBLE(mwin->video_window)
                || gtk_xine_is_fullscreen(mwin->gtx));
        GdkGeometry geom;
        
        DBG("  gtk_xine_open() returned");
        
        xfmedia_plugins_forward_signal_VOID("stream-loaded");
        
        has_video = gtk_xine_get_stream_info(mwin->gtx,
                XINE_STREAM_INFO_HAS_VIDEO);
        
        if(!xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis")
                && was_visible && !has_video)
        {
            if(gtk_xine_is_fullscreen(mwin->gtx)) {
                if(GTK_WIDGET_VISIBLE(mwin->fs_controls))
                    gtk_widget_hide(mwin->fs_controls);
                gtk_xine_set_fullscreen(mwin->gtx, FALSE);
            }
            gtk_widget_hide(mwin->video_window);
        } else if(has_video) {
            if(!was_visible)
                gtk_widget_show(mwin->video_window);
            gtk_xine_set_vis(mwin->gtx, NULL);
        } else if(xfmedia_settings_get_bool("/xfmedia/general/show_audio_vis")
                && !has_video)
        {
            gtk_widget_show(mwin->video_window);
            gtk_xine_set_vis(mwin->gtx,
                    xfmedia_settings_get_string("/xfmedia/general/selected_vis"));
        }
        
        DBG("dealt with video window");
        
        gtk_xine_set_param(mwin->gtx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 0);
        if(!gtk_xine_play(mwin->gtx, 0, 0)) {
            DBG("gtk_xine_play() failed");
            if(!was_visible)
                gtk_widget_hide(mwin->video_window);
            ret = FALSE;
            gtk_xine_set_param(mwin->gtx, XINE_PARAM_AUDIO_CLOSE_DEVICE, 1);
        } else {
            if(xfmedia_settings_get_int("/xfmedia/playlist/metadata_load_impetus") <= 1) {
                DBG("gtk_xine_play() succeeded, getting file info");
                if(!xfmedia_mainwin_get_file_info(mwin, NULL, &name, &song_len) || !name)
                    name = xfmedia_filename_to_name(filename);
                DBG("  done");
                xfmedia_playlist_set_info(mwin->plist->file_list, &itr, name, song_len);
            } else {
                gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                        &itr, PLAYLIST_COL_NAME, &name, NULL);
            }
            
            xfmedia_mainwin_set_song_label(mwin, name);
            
            if(GTK_WIDGET_VISIBLE(mwin->video_window)) {
                gtk_window_set_title(GTK_WINDOW(mwin->video_window),
                        _("Xfmedia Video"));
            }
            
            g_free(name);
            
            if(GTK_WIDGET_VISIBLE(mwin->video_window)) {
                width = gtk_xine_get_stream_info(mwin->gtx,
                        XINE_STREAM_INFO_VIDEO_WIDTH);
                height = gtk_xine_get_stream_info(mwin->gtx,
                        XINE_STREAM_INFO_VIDEO_HEIGHT);
                if(width < 1 || height < 1) {
                    width = 480;
                    height = 360;
                }
                gtk_window_resize(GTK_WINDOW(mwin->video_window),
                        width, height);
                geom.min_width = geom.min_height = 1;
                geom.min_aspect = geom.max_aspect = (gdouble)width / height;
                DBG("window size: %dx%d, aspect: %f", width, height, geom.max_aspect);
                gtk_window_set_geometry_hints(GTK_WINDOW(mwin->video_window),
                        GTK_WIDGET(mwin->gtx), &geom,
                        GDK_HINT_MIN_SIZE|GDK_HINT_ASPECT);
            }
            
            xfmedia_mainwin_disconnect_slider_callback(mwin);
            gtk_range_set_value(GTK_RANGE(mwin->position_slider), 0.0);
            gtk_range_set_value(GTK_RANGE(mwin->fs_position_slider), 0.0);
            xfmedia_mainwin_connect_slider_callback(mwin);
            
            if(gtk_xine_get_stream_info(mwin->gtx, XINE_STREAM_INFO_SEEKABLE)) {
                if(song_len == -2) {
                    gtk_tree_model_get(GTK_TREE_MODEL(mwin->plist->file_list),
                            &itr, PLAYLIST_COL_LENGTH, &song_len, -1);
                }
                mwin->plist->cur_playing_len = song_len;
                if(song_len <= 0)
                    song_len = 1;
                
                gtk_range_set_range(GTK_RANGE(mwin->position_slider), 0.0,
                        (gdouble)song_len);
                gtk_range_set_range(GTK_RANGE(mwin->fs_position_slider), 0.0,
                        (gdouble)song_len);
                gtk_widget_set_sensitive(mwin->position_slider, TRUE);
            } else
                gtk_widget_set_sensitive(mwin->position_slider, FALSE);
            
            xfmedia_mainwin_set_pause_buttons(mwin);
            
            xfmedia_playlist_shuffle_check_set(mwin->plist, row_num);
            xfmedia_plugins_forward_signal_VOID("stream-started");
            ret = TRUE;
        }
    }
    
    g_free(filename);
    
    return ret;
}

static GtkTreePath *
playlist_get_next_path(XfMediaMainwin *mwin)
{
    gint tot_items, next_id, attempts = 0, cutoff;
    GtkTreeIter itr;
    GtkTreePath *path = NULL;
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/shuffle")) {
        tot_items = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL);
        cutoff = tot_items * 0.75;
        
        do {
            if(attempts > 0)
                DBG("shuffle collision!  trying again (attempt %d)", attempts);
            
#if HAVE_SRANDOM
            next_id = random() % tot_items;
#else
            next_id = rand() % tot_items;
#endif
            if(++attempts > cutoff) {
                xfmedia_playlist_shuffle_unset_all(mwin->plist);
                attempts = 0;
            }
        } while(!xfmedia_playlist_shuffle_check(mwin->plist, next_id));
        
        path = gtk_tree_path_new_from_indices(next_id, -1);
    } else if(mwin->plist->cur_playing
            && gtk_tree_row_reference_valid(mwin->plist->cur_playing))
    {
        path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
        gtk_tree_path_next(path);
        if(!gtk_tree_model_get_iter(GTK_TREE_MODEL(mwin->plist->file_list),
                &itr, path))
        {
            /* this means we're at the end of the list (i hope) */
            if(xfmedia_settings_get_bool("/xfmedia/playlist/repeat")) {
                gtk_tree_path_free(path);
                path = gtk_tree_path_new_from_indices(0, -1);
            } else {
                gtk_xine_stop(mwin->gtx);
                if(gtk_xine_is_fullscreen(mwin->gtx))
                    gtk_xine_set_fullscreen(mwin->gtx, FALSE);
                if(GTK_WIDGET_VISIBLE(mwin->video_window))
                    gtk_widget_hide(mwin->video_window);
                if(path)
                    gtk_tree_path_free(path);
                path = NULL;
            }
        }
    } else
        path = gtk_tree_path_new_from_indices(0, -1);
    
    return path;
}

gboolean
xfmedia_playlist_play_file_at_path_persistent(XfMediaMainwin *mwin,
        GtkTreePath *path)
{
    gboolean done = FALSE, played = FALSE;
    
    if(path) {
        played = xfmedia_playlist_play_file_at_path(mwin, path);
        if(played)
            return TRUE;
    }
    
    while(!done) {
        path = playlist_get_next_path(mwin);
        if(path) {
            played = xfmedia_playlist_play_file_at_path(mwin, path);
            gtk_tree_path_free(path);
            if(played || !xfmedia_settings_get_bool("/xfmedia/playlist/auto_advance"))
                done = TRUE;
        } else
            done = TRUE;
    }
    
    return played;
}

static void
cell_render_title(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
        GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
    XfMediaMainwin *mwin = data;
    GtkTreePath *cur_path = NULL, *path = NULL;
    gboolean is_queued = FALSE;
    
    if(mwin->plist->cur_playing) {
        cur_path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
        path = gtk_tree_model_get_path(tree_model, iter);
    }
    
    gtk_tree_model_get(tree_model, iter, PLAYLIST_COL_QUEUED, &is_queued, -1);
    
    if(cur_path && path && !gtk_tree_path_compare(cur_path, path)) {
        g_object_set(G_OBJECT(cell), "weight", PANGO_WEIGHT_BOLD,
                "style-set", FALSE, NULL);
    } else if(is_queued) {
        g_object_set(G_OBJECT(cell), "style", PANGO_STYLE_ITALIC,
                "style-set", TRUE, NULL);
    } else {
        g_object_set(G_OBJECT(cell), "weight", PANGO_WEIGHT_NORMAL,
                "style-set", FALSE, NULL);
    }
    
    if(cur_path)
        gtk_tree_path_free(cur_path);
    if(cur_path)
        gtk_tree_path_free(path);
}

void
xfmedia_playlist_set_info(GtkListStore *ls, GtkTreeIter *itr, const gchar *name,
        gint length)
{
    gchar length_str[32];
    gchar *utf8 = NULL;
    
    utf8 = xfmedia_brutal_utf8_validate(name);
    
    if(length >= 0) {
        g_snprintf(length_str, 32, "%02d:%02d", length/60, length%60);
        gtk_list_store_set(ls, itr, PLAYLIST_COL_NAME, utf8?utf8:name,
                PLAYLIST_COL_LENGTH, length, PLAYLIST_COL_LENGTH_S,
                length_str, PLAYLIST_COL_INFOLOADED, TRUE, -1);
    } else
        gtk_list_store_set(ls, itr, PLAYLIST_COL_NAME, utf8?utf8:name, -1);
    
    if(utf8)
        g_free(utf8);
}

XfMediaPlaylist *
xfmedia_playlist_new(XfMediaMainwin *mwin)
{
    GtkWidget *topvbox, *sw, *hbox, *treeview, *btn, *img, *evtbox, *spacer,
        *menu, *mi;
    GdkPixbuf *pix;
    GtkListStore *ls;
    GtkTreeViewColumn *col;
    GtkCellRenderer *render;
    XfMediaPlaylist *plist;
    GtkTooltips *ttips = mwin->tooltips;
    gint iw, ih;
    const GtkTargetEntry tv_targets[] = {
        { "PLAYLIST_ITEM", GTK_TARGET_SAME_WIDGET, DROP_TGT_REORDER },
        { "text/uri-list", 0, DROP_TGT_URILIST },
        { "STRING", 0, DROP_TGT_STRING },
    };
    
    plist = g_new0(XfMediaPlaylist, 1);
    plist->shuffle_history = g_queue_new();
    
    plist->playlist_widget = topvbox = gtk_vbox_new(FALSE, BORDER);
    
    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER,
            GTK_POLICY_ALWAYS);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
            GTK_SHADOW_ETCHED_IN);
    gtk_widget_show(sw);
    gtk_box_pack_start(GTK_BOX(topvbox), sw, TRUE, TRUE, 0);
    
    plist->file_list = ls = gtk_list_store_new(PLAYLIST_N_COLS, G_TYPE_INT,
            G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT,
            G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
    
    plist->treeview = treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(ls));
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(treeview), FALSE);
    gtk_tree_view_set_reorderable(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
    
    render = gtk_cell_renderer_text_new();
    col = gtk_tree_view_column_new_with_attributes(_("#"), render, "text",
            PLAYLIST_COL_NUM, NULL);
    gtk_tree_view_column_set_resizable(col, FALSE);
    gtk_tree_view_column_set_cell_data_func(col, render, cell_render_title,
            mwin, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
    
    render = exo_cell_renderer_ellipsized_text_new();
	if(gtk_major_version == 2 && gtk_minor_version >= 6)
		g_object_set(G_OBJECT(render), "ellipsize", EXO_PANGO_ELLIPSIZE_END, NULL);
	else {
		g_object_set(G_OBJECT(render), "ellipsize", EXO_PANGO_ELLIPSIZE_END,
				"ellipsize-set", TRUE, NULL);
	}
    col = gtk_tree_view_column_new_with_attributes(_("Title"), render, "text",
            PLAYLIST_COL_NAME, NULL);
    g_object_set(G_OBJECT(col), "resizable", TRUE, "expand", TRUE, NULL);
    gtk_tree_view_column_set_cell_data_func(col, render, cell_render_title,
            mwin, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
    
    render = gtk_cell_renderer_text_new();
    col = gtk_tree_view_column_new_with_attributes(_("Length"), render, "text",
            PLAYLIST_COL_LENGTH_S, NULL);
    gtk_tree_view_column_set_resizable(col, FALSE);
    gtk_tree_view_column_set_cell_data_func(col, render, cell_render_title,
            mwin, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
    
    plist->tree_sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    gtk_tree_selection_set_mode(plist->tree_sel, GTK_SELECTION_MULTIPLE);
    gtk_tree_view_set_enable_search(GTK_TREE_VIEW(treeview), TRUE);
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(treeview), PLAYLIST_COL_NAME);
    gtk_widget_show(treeview);
    gtk_container_add(GTK_CONTAINER(sw), treeview);
    gtk_widget_add_events(treeview, GDK_BUTTON_PRESS|GDK_KEY_PRESS);
    g_signal_connect(G_OBJECT(treeview), "button-press-event",
            G_CALLBACK(xfmedia_playlist_treeview_buttonpress_cb), mwin);
    g_signal_connect(G_OBJECT(treeview), "key-press-event",
            G_CALLBACK(xfmedia_playlist_treeview_keypress_cb), mwin);
    g_signal_connect(G_OBJECT(gtk_tree_view_get_vadjustment(GTK_TREE_VIEW(treeview))),
            "value-changed", G_CALLBACK(xfmedia_playlist_treeview_scroll_cb), mwin);
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(treeview),
            GDK_BUTTON1_MASK, tv_targets, 3, GDK_ACTION_MOVE);
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(treeview), tv_targets,
            3, GDK_ACTION_COPY);
    g_signal_connect(G_OBJECT(treeview), "drag-data-received",
            G_CALLBACK(xfmedia_playlist_treeview_drag_rcv_cb), mwin);
    g_signal_connect(G_OBJECT(treeview), "drag-data-get",
            G_CALLBACK(xfmedia_playlist_treeview_drag_get_cb), mwin);
    
    evtbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(evtbox), FALSE);
    gtk_widget_show(evtbox);
    gtk_box_pack_start(GTK_BOX(topvbox), evtbox, FALSE, FALSE, 0);
    gtk_widget_add_events(evtbox, GDK_BUTTON_PRESS);
    g_signal_connect(G_OBJECT(evtbox), "button-press-event",
            G_CALLBACK(xfmedia_mainwin_menu_cb), mwin);
    
    hbox = gtk_hbox_new(FALSE, BORDER-4);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(evtbox), hbox);
    
    img = gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_toggle_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "toggled",
            G_CALLBACK(xfmedia_playlist_add_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Add entry to playlist"), NULL);
    
    img = gtk_image_new_from_stock(GTK_STOCK_REMOVE, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_remove_cb), plist);
    gtk_tooltips_set_tip(ttips, btn, _("Remove entry from playlist"), NULL);
    
    img = gtk_image_new_from_stock(GTK_STOCK_GO_UP, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_go_up_cb), plist);
    gtk_tooltips_set_tip(ttips, btn, _("Move selected playlist entry up"), NULL);
    
    img = gtk_image_new_from_stock(GTK_STOCK_GO_DOWN, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_go_down_cb), plist);
    gtk_tooltips_set_tip(ttips, btn, _("Move selected playlist entry down"), NULL);
    
    spacer = gtk_alignment_new(0, 0, 0, 0);
    gtk_widget_show(spacer);
    gtk_box_pack_start(GTK_BOX(hbox), spacer, TRUE, TRUE, 0);
    
    gtk_icon_size_lookup(GTK_ICON_SIZE_SMALL_TOOLBAR, &iw, &ih);
    
    pix = xfce_icon_theme_load(icon_theme, "media-shuffle", iw);
    if(!pix)
        pix = xfce_icon_theme_load(icon_theme, "stock_media-shuffle", iw);
    if(!pix)
        pix = xfce_icon_theme_load(icon_theme, "xfmedia-shuffle", iw);
    if(pix) {
        img = gtk_image_new_from_pixbuf(pix);
        gtk_widget_show(img);
        g_object_unref(G_OBJECT(pix));
        plist->shuffle_btn = btn = gtk_toggle_button_new();
        gtk_container_add(GTK_CONTAINER(btn), img);
    } else
        plist->shuffle_btn = btn = gtk_toggle_button_new_with_label("S");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(btn),
            xfmedia_settings_get_bool("/xfmedia/playlist/shuffle"));
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "toggled",
            G_CALLBACK(xfmedia_playlist_shuffle_toggle_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Randomize play order"), NULL);
    
    pix = xfce_icon_theme_load(icon_theme, "repeat", iw);
    if(!pix)
        pix = xfce_icon_theme_load(icon_theme, "stock_repeat", iw);
    if(!pix)
        pix = xfce_icon_theme_load(icon_theme, "xfmedia-repeat", iw);
    if(pix) {
        img = gtk_image_new_from_pixbuf(pix);
        gtk_widget_show(img);
        g_object_unref(G_OBJECT(pix));
        plist->repeat_btn = btn = gtk_toggle_button_new();
        gtk_container_add(GTK_CONTAINER(btn), img);
    } else
        plist->repeat_btn = btn = gtk_toggle_button_new_with_label("R");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(btn),
            xfmedia_settings_get_bool("/xfmedia/playlist/repeat"));
    gtk_widget_show(btn);
    gtk_box_pack_start(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "toggled",
            G_CALLBACK(xfmedia_playlist_repeat_toggle_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Repeat after playlist finishes"), NULL);
    
    spacer = gtk_alignment_new(0, 0, 0, 0);
    //gtk_widget_set_size_request(spacer, 25, -1);
    gtk_widget_show(spacer);
    gtk_box_pack_start(GTK_BOX(hbox), spacer, TRUE, TRUE, 0);
    
    img = gtk_image_new_from_stock(GTK_STOCK_SAVE_AS, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_end(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_save_as_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Save current playlist"), NULL);
    
    img = gtk_image_new_from_stock(GTK_STOCK_OPEN, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_end(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_open_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Open a saved playlist"), NULL);
    
    img = gtk_image_new_from_stock(GTK_STOCK_NEW, GTK_ICON_SIZE_SMALL_TOOLBAR);
    gtk_widget_show(img);
    btn = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(btn), img);
    gtk_widget_show(btn);
    gtk_box_pack_end(GTK_BOX(hbox), btn, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(btn), "clicked",
            G_CALLBACK(xfmedia_playlist_new_cb), mwin);
    gtk_tooltips_set_tip(ttips, btn, _("Start a new playlist"), NULL);
    
    plist->popup_menu = menu = gtk_menu_new();
    gtk_widget_show(menu);
    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &iw, &ih);
    
    img = gtk_image_new_from_stock(GTK_STOCK_DIALOG_INFO, GTK_ICON_SIZE_MENU);
    gtk_widget_show(img);
    mi = gtk_image_menu_item_new_with_mnemonic(_("_File Info"));
    gtk_widget_show(mi);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_file_info_cb), mwin);
    
    mi = gtk_separator_menu_item_new();
    gtk_widget_show(mi);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    
    img = gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_MENU);
    gtk_widget_show(mi);
    mi = gtk_image_menu_item_new_with_mnemonic(_("Add _Mediamark..."));
    gtk_widget_show(mi);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_add_mm_cb), mwin);
    
    img = gtk_image_new_from_stock(GTK_STOCK_REMOVE, GTK_ICON_SIZE_MENU);
    gtk_widget_show(mi);
    mi = gtk_image_menu_item_new_with_mnemonic(_("_Remove Entry"));
    gtk_widget_show(mi);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_remove_cb), plist);
    
    img = gtk_image_new_from_stock(GTK_STOCK_GO_UP, GTK_ICON_SIZE_MENU);
    gtk_widget_show(mi);
    mi = gtk_image_menu_item_new_with_mnemonic(_("Move _Up"));
    gtk_widget_show(mi);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_go_up_cb), plist);
    
    img = gtk_image_new_from_stock(GTK_STOCK_GO_DOWN, GTK_ICON_SIZE_MENU);
    gtk_widget_show(mi);
    mi = gtk_image_menu_item_new_with_mnemonic(_("Move _Down"));
    gtk_widget_show(mi);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), img);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
    g_signal_connect(G_OBJECT(mi), "activate",
            G_CALLBACK(xfmedia_playlist_go_down_cb), plist);
    
    return plist;
}

void
xfmedia_playlist_renumber_file_list(GtkListStore *ls, gint from_index,
        gint to_index)
{
    GtkTreeIter itr;
    gint i, nrows;
    
    DBG("renumbering from %d to %d", from_index, to_index);
    
    if(from_index < 0)
        from_index = 0;
    i = from_index + 1;
    nrows = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(ls), NULL);
    if(to_index >= nrows || to_index < 0)
        to_index = nrows - 1;
    
    if(from_index == 0 && gtk_tree_model_get_iter_first(GTK_TREE_MODEL(ls), &itr)) {
        gtk_list_store_set(ls, &itr, PLAYLIST_COL_NUM, 1, -1);
        i++;
    } else if(from_index == 0 || (from_index > 0
            && !gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(ls), &itr,
                                             NULL, from_index-1)))
    {
        return;
    }
    
    for(; gtk_tree_model_iter_next(GTK_TREE_MODEL(ls), &itr); i++) {
        gtk_list_store_set(ls, &itr, PLAYLIST_COL_NUM, i, -1);
        if(i == to_index + 1)
            break;
    }
}

void
xfmedia_playlist_add_file(XfMediaPlaylist *plist, gint index, const gchar *name,
        gint length, const gchar *filename, gboolean info_loaded)
{
    GtkTreeIter itr;
    gchar length_str[32], *utf8 = NULL;
    
    if(index < 0)
        index = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(plist->file_list), NULL);
    
    gtk_list_store_insert(plist->file_list, &itr, index);
    
    if(length >= 0)
        g_snprintf(length_str, 32, "%02d:%02d", length/60, length%60);
    
    utf8 = xfmedia_brutal_utf8_validate(name);
    
    gtk_list_store_set(plist->file_list, &itr,
            PLAYLIST_COL_NUM, index + 1,
            PLAYLIST_COL_NAME, utf8 ? utf8 : name,
            PLAYLIST_COL_LENGTH, length,
            PLAYLIST_COL_FILENAME, filename,
            PLAYLIST_COL_LENGTH_S, (length >= 0) ? length_str : "??:??",
            PLAYLIST_COL_INFOLOADED, info_loaded,
            PLAYLIST_COL_QUEUED, FALSE,
            -1);
    
    if(utf8)
        g_free(utf8);
}

gboolean
xfmedia_playlist_remove_file(XfMediaPlaylist *plist, GtkTreePath *path,
        gint index)
{
    gchar *filename = NULL;
    GtkTreePath *path1 = NULL;
    GtkTreeIter itr;
    gboolean ret = FALSE;
    
    if(path)
        index = xfmedia_tree_path_to_index(path);
    else if(index >= 0)
        path1 = path = gtk_tree_path_new_from_indices(index, -1);
    
    if(gtk_tree_model_get_iter(GTK_TREE_MODEL(plist->file_list), &itr, path)) {
        gtk_tree_model_get(GTK_TREE_MODEL(plist->file_list), &itr,
                PLAYLIST_COL_FILENAME, &filename, -1);
    }
    
    if(xfmedia_list_store_remove_path(plist->file_list, path)) {
        plist->is_dirty = TRUE;
        plist->save_on_exit = TRUE;
        ret = TRUE;
    }
    
    if(filename) {
        xfmedia_plugins_forward_signal_STRING("playlist-entry-removed",
                filename);
        g_free(filename);
    }
    
    if(path1)
        gtk_tree_path_free(path1);
    
    return ret;
}

G_INLINE_FUNC gboolean
file_needs_dir(const gchar *filename)
{
    g_return_val_if_fail(filename && *filename, FALSE);
    
    return (!strstr(filename, ":/") && *filename != '/');
}

G_INLINE_FUNC gchar *
string_strip_newlines(gchar *str)
{
    while(str[strlen(str)-1] == '\n' || str[strlen(str)-1] == '\r')
        str[strlen(str)-1] = 0;
    
    return str;
}

static gint
xfmedia_playlist_load_flat_file(FILE *fp, const gchar *basedir,
        XfMediaMainwin *mwin)
{
    gchar buf[PATH_MAX], buf1[PATH_MAX];
    gint nadded = 0;
    
    while(fgets(buf, PATH_MAX, fp)) {
        if(!*buf || *buf == '#' || *buf == '\n' || *buf == '\r')
            continue;
        
        if(file_needs_dir(buf)) {
            g_snprintf(buf1, PATH_MAX, "%s/%s", basedir, buf);
            xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf1), -1, TRUE);
        } else
            xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf), -1, TRUE);
        nadded++;
    }
    
    return nadded;
}

static gint
xfmedia_playlist_load_m3u(FILE *fp, const gchar *basedir, XfMediaMainwin *mwin)
{
    gchar buf[PATH_MAX], buf1[PATH_MAX], buf2[PATH_MAX], *p, *q;
    gint song_len, nadded = 0;
    
    /* check header */
    if(fgets(buf, PATH_MAX, fp)) {
        if(g_ascii_strncasecmp(buf, "#EXTM3U", 7)) {
            /* not an M3U file */
            if(file_needs_dir(buf)) {
                g_snprintf(buf1, PATH_MAX, "%s/%s", basedir, buf);
                xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf1), -1, TRUE);
            } else
                xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf), -1, TRUE);
            
            /* fall back to flat file loading */
            return xfmedia_playlist_load_flat_file(fp, basedir, mwin) + 1;
        }
    }
    
    while(fgets(buf, PATH_MAX, fp)) {
        if(!g_ascii_strncasecmp(buf, "#EXTINF:", 8)) {
            p = buf + 8;
            q = strstr(p, ",");
            if(!p || *q == 0)
                continue;
            *q = 0;
            q++;
            song_len = atoi(p);
            if(song_len < 0 || *q == 0)
                continue;
            if(fgets(buf1, PATH_MAX, fp)) {
                if(file_needs_dir(buf1)) {
                    g_snprintf(buf2, PATH_MAX, "%s/%s", basedir, buf1);
                    xfmedia_playlist_add_file(mwin->plist, -1,
                            string_strip_newlines(q), song_len,
                            string_strip_newlines(buf2), TRUE);
                } else {
                    xfmedia_playlist_add_file(mwin->plist, -1,
                            string_strip_newlines(q), song_len,
                            string_strip_newlines(buf1), TRUE);
                }
                nadded++;
            }
        } else {
            if(file_needs_dir(buf)) {
                g_snprintf(buf1, PATH_MAX, "%s/%s", basedir, buf);
                xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf1), -1, TRUE);
            } else
                xfmedia_mainwin_add_file(mwin, string_strip_newlines(buf), -1, TRUE);
            nadded++;
        }
    }
    
    return nadded;
}

static gint
xfmedia_playlist_load_pls(const gchar *filename, const gchar *basedir,
        XfMediaMainwin *mwin)
{
    XfceRc *rcfile;
    gint i, nentries, length;
    const gchar *music_file, *title, *length_str;
    gchar key[128], buf[PATH_MAX];
    gint nadded = 0;
    
    rcfile = xfce_rc_simple_open(filename, TRUE);
    if(xfce_rc_has_group(rcfile, "playlist")) {
        xfce_rc_set_group(rcfile, "playlist");
        
        nentries = xfce_rc_read_int_entry(rcfile, "numberofentries", 0);
        for(i = 1; i <= nentries; i++) {
            g_snprintf(key, 128, "File%d", i);
            music_file = xfce_rc_read_entry(rcfile, key, NULL);
            if(!music_file)
                continue;
            
            g_snprintf(key, 128, "Title%d", i);
            title = xfce_rc_read_entry(rcfile, key, NULL);
            g_snprintf(key, 128, "Length%d", i);
            length_str = xfce_rc_read_entry(rcfile, key, NULL);
            if(length_str)
                length = atoi(length_str);
            
            if(file_needs_dir(music_file))
                g_snprintf(buf, PATH_MAX, "%s/%s", basedir, music_file);
            else
                *buf = 0;
            
            /* set song data directly, and only set info_loaded to TRUE if
             * we got both a length and title from the file */
            xfmedia_playlist_add_file(mwin->plist, -1, title,
                    length_str ? length : 0, *buf ? buf : music_file,
                    (title && length_str) ? TRUE : FALSE);
            nadded += 1;
        }
    }
    xfce_rc_close(rcfile);
    
    return nadded;
}

gboolean
xfmedia_playlist_load(const gchar *filename, XfMediaMainwin *mwin,
        gboolean clear_first)
{
    FILE *fp;
    gchar *p, *basedir;
    gboolean ret = FALSE, is_m3u = FALSE, is_pls = FALSE;
    gint nadded = 0;
    
    if(g_str_has_suffix(filename, ".m3u") || g_str_has_suffix(filename, ".M3U"))
        is_m3u = TRUE;
    else if(g_str_has_suffix(filename, ".pls") || g_str_has_suffix(filename, ".PLS"))
        is_pls = TRUE;
    
    fp = fopen(filename, "r");
    if(fp) {
        if(clear_first)
            gtk_list_store_clear(mwin->plist->file_list);
        
        p = g_strrstr(filename, "/");
        if(p)
            basedir = g_strndup(filename, strlen(filename)-strlen(p));
        else
            basedir = g_strdup(g_get_current_dir());
        
        if(is_m3u)
            nadded = xfmedia_playlist_load_m3u(fp, basedir, mwin);
        else if(is_pls) {
            fclose(fp);
            fp = NULL;
            nadded = xfmedia_playlist_load_pls(filename, basedir, mwin);
        } else
            nadded = xfmedia_playlist_load_flat_file(fp, basedir, mwin);
        
        g_free(basedir);
        
        if(fp)
            fclose(fp);
        ret = TRUE;
        
        if(clear_first)
            xfmedia_playlist_shuffle_init_entries(mwin->plist, nadded);
        else
            xfmedia_playlist_shuffle_add_entries(mwin->plist, -1, nadded);
    }
    
    return ret;
}

static void
xfmedia_playlist_save_flat_file(FILE *fp, GtkListStore *ls)
{
    GtkTreeIter itr;
    gchar *music_file = NULL;
    
    if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(ls), &itr)) {
        do {
            gtk_tree_model_get(GTK_TREE_MODEL(ls), &itr,
                    PLAYLIST_COL_FILENAME, &music_file, -1);
            
            if(music_file) {
                fprintf(fp, "%s\n", music_file);
                g_free(music_file);
                music_file = NULL;
            }
        } while(gtk_tree_model_iter_next(GTK_TREE_MODEL(ls), &itr));
    }
}

static void
xfmedia_playlist_save_m3u(FILE *fp, GtkListStore *ls)
{
    GtkTreeIter itr;
    gchar *name = NULL, *music_file = NULL;
    gint song_len = 0;
    gboolean info_loaded = FALSE;
    
    if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(ls), &itr)) {
        fputs("#EXTM3U\n", fp);
        
        do {
            gtk_tree_model_get(GTK_TREE_MODEL(ls), &itr,
                    PLAYLIST_COL_NAME, &name, PLAYLIST_COL_FILENAME,
                    &music_file, PLAYLIST_COL_LENGTH, &song_len,
                    PLAYLIST_COL_INFOLOADED, &info_loaded, -1);
            
            if(music_file && info_loaded) {
                fprintf(fp, "#EXTINF:%d,%s\n%s\n", song_len,
                        name ? name : "", music_file);
            } else if(music_file)
                fprintf(fp, "%s\n", music_file);
            
            if(music_file)
                g_free(music_file);
            if(name)
                g_free(name);
            
            name = music_file = NULL;
            song_len = 0;
            info_loaded = FALSE;
        } while(gtk_tree_model_iter_next(GTK_TREE_MODEL(ls), &itr));
    }
}

static void
xfmedia_playlist_save_pls(FILE *fp, GtkListStore *ls)
{
    GtkTreeIter itr;
    gchar *name = NULL, *music_file = NULL;
    gint nentries, i = 1, song_len = 0;
    gboolean info_loaded = FALSE;
    
    if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(ls), &itr)) {
        nentries = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(ls), NULL);
        fprintf(fp, "[playlist]\nnumberofentries=%d\n", nentries);
        
        do {
            gtk_tree_model_get(GTK_TREE_MODEL(ls), &itr,
                    PLAYLIST_COL_NAME, &name, PLAYLIST_COL_FILENAME,
                    &music_file, PLAYLIST_COL_LENGTH, &song_len,
                    PLAYLIST_COL_INFOLOADED, &info_loaded, -1);
            
            if(music_file && info_loaded) {
                fprintf(fp, "File%d=%s\nTitle%d=%s\nLength%d=%d\n",
                        i, music_file, i, (name ? name : ""), i, song_len);
            } else
                fprintf(fp, "File%d=%s\n", i, music_file ? music_file : "");
            
            i++;
            
            if(music_file)
                g_free(music_file);
            if(name)
                g_free(name);
            
            name = music_file = NULL;
            song_len = 0;
            info_loaded = FALSE;
        } while(gtk_tree_model_iter_next(GTK_TREE_MODEL(ls), &itr));
    }
}

gboolean
xfmedia_playlist_save(GtkListStore *ls, const gchar *filename)
{
    FILE *fp;
    gboolean ret = FALSE, is_m3u = FALSE, is_pls = FALSE;
    
    if(g_str_has_suffix(filename, ".m3u") || g_str_has_suffix(filename, ".M3U"))
        is_m3u = TRUE;
    else if(g_str_has_suffix(filename, ".pls") || g_str_has_suffix(filename, ".PLS"))
        is_pls = TRUE;
    
    fp = fopen(filename, "w");
    if(fp) {
        if(is_m3u)
            xfmedia_playlist_save_m3u(fp, ls);
        else if(is_pls)
            xfmedia_playlist_save_pls(fp, ls);
        else
            xfmedia_playlist_save_flat_file(fp, ls);
        
        fclose(fp);
        ret = TRUE;
    }
    
    return ret;
}

void
xfmedia_playlist_destroy(XfMediaPlaylist *plist)
{
    gint tot, i;
    
    if(plist->cur_playing)
        gtk_tree_row_reference_free(plist->cur_playing);
    if(plist->chooser_last_dir)
        g_free(plist->chooser_last_dir);
    
    tot = g_queue_get_length(plist->shuffle_history);
    for(i = tot-1; i >= 0; i--)
        gtk_tree_row_reference_free((GtkTreeRowReference *)g_queue_pop_nth(plist->shuffle_history, i));
    g_queue_free(plist->shuffle_history);
    
    g_free(plist);
}

gboolean
xfmedia_playlist_is_playlist_file(const gchar *filename)
{
    return (g_str_has_suffix(filename, ".m3u")
            || g_str_has_suffix(filename, ".pls")
            || g_str_has_suffix(filename, ".M3U")
            || g_str_has_suffix(filename, ".PLS"));
}


/*
 * playlist shuffle-play management
 */

gboolean
xfmedia_playlist_shuffle_check(XfMediaPlaylist *plist, guint index)
{
    g_return_val_if_fail(plist->shuffle_bitmap
            && index < plist->shuffle_bitmap_len, FALSE);
    
    return (plist->shuffle_bitmap[index] == 0x0);
}

gboolean
xfmedia_playlist_shuffle_check_set(XfMediaPlaylist *plist, guint index)
{
    g_return_val_if_fail(plist->shuffle_bitmap
            && index < plist->shuffle_bitmap_len, FALSE);
    
    if(plist->shuffle_bitmap[index])
        return FALSE;
    
    return (plist->shuffle_bitmap[index] = 0x1);  /* yes, a single '=' is correct */
}

void
xfmedia_playlist_shuffle_unset(XfMediaPlaylist *plist, guint index)
{
    g_return_if_fail(index < plist->shuffle_bitmap_len);
    
    plist->shuffle_bitmap[index] = 0x0;
}
    
void
xfmedia_playlist_shuffle_unset_all(XfMediaPlaylist *plist)
{
    if(!plist->shuffle_bitmap)
        return;
    
    memset(plist->shuffle_bitmap, 0x0, plist->shuffle_bitmap_len);
}

void
xfmedia_playlist_shuffle_add_entries(XfMediaPlaylist *plist, gint start,
        guint len)
{
    g_return_if_fail(len > 0);
    
    plist->shuffle_bitmap = g_realloc(plist->shuffle_bitmap,
            plist->shuffle_bitmap_len + len);
    
    if(start == -1 || start >= plist->shuffle_bitmap_len)
        memset(plist->shuffle_bitmap+plist->shuffle_bitmap_len, 0x0, len);
    else {
        memmove(plist->shuffle_bitmap+start+len, plist->shuffle_bitmap+start, len);
        memset(plist->shuffle_bitmap+start, 0x0, len);
    }
        
    plist->shuffle_bitmap_len += len;
}

void
xfmedia_playlist_shuffle_remove_entries(XfMediaPlaylist *plist, guint start,
        gint len)
{
    g_return_if_fail(len != 0 && start < plist->shuffle_bitmap_len);
    
    if(len < 0)
        len = plist->shuffle_bitmap_len - start;
    
    if(len == plist->shuffle_bitmap_len) {
        g_free(plist->shuffle_bitmap);
        plist->shuffle_bitmap = NULL;
    } else {
        memmove(plist->shuffle_bitmap+start, plist->shuffle_bitmap+start+len, len);
        plist->shuffle_bitmap = g_realloc(plist->shuffle_bitmap,
                plist->shuffle_bitmap_len - len);
    }
    
    plist->shuffle_bitmap_len -= len;
}

void
xfmedia_playlist_shuffle_swap_entries(XfMediaPlaylist *plist, guint a, guint b)
{
    guint8 tmp;
    
    g_return_if_fail(a < plist->shuffle_bitmap_len
            && b < plist->shuffle_bitmap_len && a != b);
    
    tmp = plist->shuffle_bitmap[a];
    plist->shuffle_bitmap[a] = plist->shuffle_bitmap[b];
    plist->shuffle_bitmap[b] = tmp;
}

void
xfmedia_playlist_shuffle_init_entries(XfMediaPlaylist *plist, guint len)
{
    if(len == 0) {
        if(plist->shuffle_bitmap) {
            g_free(plist->shuffle_bitmap);
            plist->shuffle_bitmap = NULL;
            plist->shuffle_bitmap_len = 0;
        }
    } else {
        plist->shuffle_bitmap = g_realloc(plist->shuffle_bitmap, len);
        memset(plist->shuffle_bitmap, 0x0, len);
        plist->shuffle_bitmap_len = len;
    }
}
