/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _XFMEDIA__PLAYLIST_H_
#define _XFMEDIA__PLAYLIST_H_

#include "xfmedia-common.h"

G_BEGIN_DECLS

XfMediaPlaylist *xfmedia_playlist_new(XfMediaMainwin *mwin);
void xfmedia_playlist_destroy(XfMediaPlaylist *plist);

void xfmedia_playlist_add_file(XfMediaPlaylist *plist, gint index, const gchar *name, gint length, const gchar *filename, gboolean info_loaded);
gboolean xfmedia_playlist_remove_file(XfMediaPlaylist *plist, GtkTreePath *path, gint index);

gboolean xfmedia_playlist_play_file_at_path(XfMediaMainwin *mwin, GtkTreePath *path);
gboolean xfmedia_playlist_play_file_at_path_persistent(XfMediaMainwin *mwin, GtkTreePath *path);

void xfmedia_playlist_set_info(GtkListStore *ls, GtkTreeIter *itr, const gchar *name, gint length);
void xfmedia_playlist_renumber_file_list(GtkListStore *ls, gint from_index, gint to_index);

gboolean xfmedia_playlist_is_playlist_file(const gchar *filename);
gboolean xfmedia_playlist_load(const gchar *filename, XfMediaMainwin *mwin, gboolean clear_first);
gboolean xfmedia_playlist_save(GtkListStore *ls, const gchar *filename);

/* playlist shuffle-play management */
gboolean xfmedia_playlist_shuffle_check(XfMediaPlaylist *plist, guint index);
gboolean xfmedia_playlist_shuffle_check_set(XfMediaPlaylist *plist, guint index);
void xfmedia_playlist_shuffle_unset(XfMediaPlaylist *plist, guint index);
void xfmedia_playlist_shuffle_unset_all(XfMediaPlaylist *plist);
void xfmedia_playlist_shuffle_add_entries(XfMediaPlaylist *plist, gint start, guint len);
void xfmedia_playlist_shuffle_remove_entries(XfMediaPlaylist *plist, guint start, gint len);
void xfmedia_playlist_shuffle_swap_entries(XfMediaPlaylist *plist, guint a, guint b);
void xfmedia_playlist_shuffle_init_entries(XfMediaPlaylist *plist, guint len);

G_END_DECLS

#endif
