/*
 *  xfmusic4-cli - simple cmdline test implementation using libxfmusic4
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>

#include <libxfce4util/i18n.h>
#include <libxfmusic4/streams.h>

#include "cli-misc.h"

void
version()
{
    g_printerr("%s v%s (alpha)\n", PACKAGE, VERSION);
    g_printerr("  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>\n");
    g_printerr("\n");
    g_printerr(_("This software is released under the terms of the GNU General Public License.  It\n"));
    g_printerr(_("is distributed WITHOUT ANY WARRANTY; without even the implied warranty of\n"));
    g_printerr(_("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  For more details, see the\n"));
    g_printerr(_("COPYING file included with the source distribution.\n\n"));
}

void
usage()
{
    g_printerr(_("Usage: %s [options] file ...\n\n"), g_get_prgname());
    g_printerr(_("General Options:\n"));
    g_printerr(_("  -v, --verbose             prints file information during playback\n"));
    g_printerr(_("  -q, --quiet               do not print anything (except errors)\n"));
    g_printerr(_("  -h, --help                prints this help message\n"));
    g_printerr(_("  -V, --version             prints version information\n"));
    g_printerr("\n");
    g_printerr(_("  -i, --input=TYPE          force the use of a specific input plugin (required\n"));
    g_printerr(_("                            when reading from stdin)\n"));
    g_printerr(_("  -o, --output=TYPE[:FILE]  use audio output driver TYPE; some drivers also\n"));
    g_printerr(_("                            require a FILE as well; available drivers are:\n"));
    g_printerr("                             ");
#ifdef HAVE_ALSA
    g_printerr(" alsa");
#endif
#ifdef HAVE_ESD
    g_printerr(" esd");
#endif
#ifdef HAVE_LIBAO
    g_printerr(" libao");
#endif
#ifdef HAVE_SNDFILE
    g_printerr(" wav");
#endif
    g_printerr("\n\n");
    g_printerr(_("Controls:\n"));
    g_printerr(_("  'q'             quits %s\n"), PACKAGE);
    g_printerr(_("  'p'             pauses or resumes playback\n"));
    g_printerr(_("  'f'             advances to the next song in the playlist\n"));
    g_printerr(_("  'b'             returns to the previous song in the playlist\n"));
    g_printerr(_("  <left>,<right>  moves backward or forward 5 seconds\n"));
    g_printerr(_("  <down>,<up>     moves backward or forward 30 seconds\n"));
    g_printerr(_("  <pgdn>,<pgup>   moves backward or forward 60 seconds\n"));
    g_printerr("\n");
}

void
print_stream_info(XfMusic *xfmusic)
{
    XfMusicStreamInfo *sinfo = xfmusic_get_stream_info(xfmusic);
    
    if(!sinfo)
        return;
    
    g_print("\n");
    if(sinfo->artist)
        g_print(_("Artist: %s\n"), sinfo->artist);
    if(sinfo->title)
        g_print(_("Title:  %s\n"), sinfo->title);
    if(sinfo->album)
        g_print(_("Album:  %s\n"), sinfo->album);
    if(sinfo->genre) {
        g_print(_("Genre:  %s"), sinfo->genre);
        if(sinfo->year != 0 || sinfo->track != 0)
            g_print(" | ");
    }
    if(sinfo->track != 0) {
        g_print(_("Track:  %02d"), sinfo->track);
        if(sinfo->year != 0)
            g_print(" | ");
    }
    if(sinfo->year != 0)
        g_print(_("Year:   %4d"), sinfo->year);
    if(sinfo->genre || sinfo->track != 0 || sinfo->year != 0)
        g_print("\n");
    if(sinfo->comment)
        g_print(_("Comment:\n%s\n"), sinfo->comment);
}
